/*
 *  File: rvp.c - Gaim RVP plug-in implementation.
 *  Copyright (C) 2003  Weihua Sun (weihua@lucent.com)
 *  Updated 2005 Waider <waider@waider.ie>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define GAIM_PLUGINS /* this switches the GAIM_INIT_PLUGIN declaration
                        from in-line build to outboard build */

#include "rvp.h"

#include <netdb.h>
#include <iconv.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#ifdef HAVE_ARPA_NAMESER_COMPAT_H
# include <arpa/nameser_compat.h>
#endif
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include "random.h"

/* gaim includes */
#include <accountopt.h>
#include <conversation.h>
#include <debug.h>
#include <ft.h>
#include <gtkutils.h>
#include <internal.h> /* this is here for the gettext bits */
#include <notify.h>
#include <plugin.h>
#include <pluginpref.h>
#include <privacy.h>
#include <server.h>
#include <util.h>
#include <version.h>

/* Let's do the compatibility dance! */
#if GAIM_MAJOR_VERSION < 2
/* gack. they dropped this. the digest-auth code is incomplete anyway */
#include <md5.h>

/* this is actually pretty sane */
#define GAIM_CONV_TYPE_CHAT GAIM_CONV_CHAT
#else
/* put here things for Gaim 2+ */
#endif

/* and this is because I'm bad, bad, bad */
#include <gtk/gtkmain.h>

#ifdef DEBUG
#include <libxml/debugXML.h>
#endif

static char HEX_ELM[17] = { '0','1','2','3','4','5','6','7',
                            '8','9','A','B','C','D','E','F','\0' };

static away_t awaymsgs[] = {
  { "online",    _("Online") },
  { "busy",      _("Busy") },
  { "idle",      _("Idle") },
  { "back-soon", _("Be Right Back") },
  { "away",      _("Away From Computer") },
  { "on-phone",  _("On The Phone") },
  { "at-lunch",  _("Out To Lunch") },
  { "offline",   _("Offline") }, /* yeah yeah */
  { "offline",   _("Appear Offline") },
  { NULL, NULL },
};

static int rvp_send_notify( GaimConnection *, const char *, gint, gchar *,
                            void * );
static void rvp_send_request( GaimConnection *, gchar *, gchar ** );
static void rvp_add_buddy( GaimConnection *, GaimBuddy *, GaimGroup * );
static void rvp_set_buddy_status( GaimConnection *, gchar *, char * );
static const char *rvp_normalize( const GaimAccount *, const char * );
static void rvp_send_request( GaimConnection *, gchar *, gchar ** );
static int rvp_bind_local_socket( GaimConnection *, int, struct sockaddr_in *,
                                  int );

/* async url stuff */
static void rvp_async_data( void *, GaimFetchUrlData *, size_t );
static void url_fetched_cb( gpointer, gint, GaimInputCondition );
static void destroy_fetch_url_data( GaimFetchUrlData * );

static void rvp_add_permit(GaimConnection *, const char *);

void rvp_chat_user_left( GaimConnection *, int, GaimBuddy * );

/*
 * This horror is just a temporary hack until I get around to redoing
 * it properly... right now I just want the SRV record ASAP.
 *
 * correct version:
 * 1. asynchronous, if possible. see the main gaim source
 *    1a. need to handle callbacks for this, which complicates code
 *        elsewhere
 * 2. need to keep track of use of SRV records and balance them
 *    accordingly. This is the intention of the unused 'last' param.
 * 3. cache what we've learned and short circuit responses. NB TTL!
 */
/* record type: 2 bytes
   record class: 2 bytes
   TTL: 4 bytes
   data length: 2 bytes */
typedef struct _rechdr {
  guint16 type;
  guint16 class;
  guint32 ttl;
  guint16 length;
} rechdr;

/* SRV record:
   priority: 2 bytes
   weight: 2 bytes
   port: 2 bytes
   target: remainder
*/
typedef struct _srv {
  guint16 priority;
  guint16 weight;
  guint16 port;
  guchar *target;
} srv_t;

/* cache for SRV records */
static GHashTable *srvcache = NULL;

srvrec *gethostbysrv( char *srvname, srvrec *last ) {
  HEADER *hdr;
  unsigned char answer[1024];
  int len = 0;
  char name[1024];
  srvrec *retval = NULL;

  if ( srvcache == NULL ) {
    gaim_debug_misc( __FUNCTION__, "Created SRV cache\n" );
    srvcache = g_hash_table_new( g_str_hash, g_str_equal );
  }

  if (( retval = (srvrec *)g_hash_table_lookup( srvcache, srvname ))
      != NULL ) {
    if ( retval->expiry > time( NULL )) {
      return retval;
    } else {
      gaim_debug_misc( __FUNCTION__, "cached %s expired\n", srvname );
      retval = NULL;
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "%s not in SRV cache\n", srvname );
  }

  bzero( answer, 1024 );
  bzero( name, 1024 );

  _res.options |= RES_DEBUG;
  if (( len = res_search( srvname, C_IN, T_SRV, answer, sizeof( answer )))
      != -1 ) {
    unsigned char *blob;
    int i, l;

    hdr = (HEADER *)answer;
    if ( ntohs( hdr->ancount ) == 0 ) {
      gaim_debug_warning( __FUNCTION__, "res_search: no records found\n" );
      goto out;
    }

    blob = &answer[HFIXEDSZ];

    for ( i = 0; i < ntohs( hdr->qdcount ); i++ ) {
      l = dn_expand( answer, answer + len, blob, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (1)\n" );
        goto out;
      }
      blob += l + QFIXEDSZ;
    }

    for ( i = 0; i < ntohs( hdr->ancount ); i++ ) {
      rechdr *header;
      srv_t *server;

      l = dn_expand( answer, answer + len, blob, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (2)\n" );
        goto out;
      }

      blob += l;

      header = (rechdr *)&blob[0];
      server = (srv_t *)&blob[10]; /* errr. magic number. */

      l = dn_expand( answer, answer + len, blob + 16, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (3)\n" );
        goto out;
      }

      retval = g_new0( srvrec, 1 );
      retval->host = g_strdup( name );
      retval->port = ntohs( server->port );

      retval->expiry = time( NULL ) + ntohl( header->ttl );
    }
  } else {
    gaim_debug_error( __FUNCTION__, "res_search for %s failed (%s)\n",
                      srvname, strerror( errno ));
  }

 out:
  if ( retval == NULL ) {
    retval = g_new0( srvrec, 1 );
    retval->host = NULL;
    retval->port = 0;
    retval->expiry = time( NULL ) + 86400; /* xxx */
  }

  g_hash_table_replace( srvcache, g_strdup( srvname ), retval );

  return retval;
}
/* end of horror */

/*
 * debug function to dump out a buddy structure
 */
static void rvp_dump_buddy( GaimBuddy *buddy ) {
  if ( buddy != NULL ) {
    RVPBuddy *rbud = buddy->proto_data;

    gaim_debug_misc( __FUNCTION__,     "ptr       %p (rvp @ %p)\n", buddy,
                     rbud );
    gaim_debug_misc( __FUNCTION__,     "buddy     %s\n", buddy->name );
    if ( buddy->alias != NULL )
      gaim_debug_misc( __FUNCTION__,   "alias     %s\n", buddy->alias );
    if ( buddy->server_alias != NULL )
      gaim_debug_misc( __FUNCTION__,   "srvalias  %s\n", buddy->server_alias );

    if ( rbud != NULL ) {
      if ( rbud->email != NULL )
        gaim_debug_misc( __FUNCTION__, "email     %s\n", rbud->email );
      if ( rbud->principal != NULL )
        gaim_debug_misc( __FUNCTION__, "principal %s\n", rbud->principal );
      if ( rbud->sessionid != NULL )
        gaim_debug_misc( __FUNCTION__, "sessionid %s\n", rbud->sessionid );
      if ( rbud->subs_id )
        gaim_debug_misc( __FUNCTION__, "subs id   %d\n", rbud->subs_id );

      if ( rbud->acl.credentials.assertion )
        gaim_debug_misc( __FUNCTION__, "CRED: assertion\n" );
      if ( rbud->acl.credentials.digest )
        gaim_debug_misc( __FUNCTION__, "CRED: digest\n" );
      if ( rbud->acl.credentials.ntlm )
        gaim_debug_misc( __FUNCTION__, "CRED: ntlm\n" );

      if ( rbud->acl.list )
        gaim_debug_misc( __FUNCTION__, "ACL: list\n" );
      if ( rbud->acl.read )
        gaim_debug_misc( __FUNCTION__, "ACL: read\n" );
      if ( rbud->acl.write )
        gaim_debug_misc( __FUNCTION__, "ACL: write\n" );
      if ( rbud->acl.send_to )
        gaim_debug_misc( __FUNCTION__, "ACL: send-to\n" );
      if ( rbud->acl.receive_from )
        gaim_debug_misc( __FUNCTION__, "ACL: receive-from\n" );
      if ( rbud->acl.readacl )
        gaim_debug_misc( __FUNCTION__, "ACL: readacl\n" );
      if ( rbud->acl.writeacl )
        gaim_debug_misc( __FUNCTION__, "ACL: writeacl\n" );
      if ( rbud->acl.presence )
        gaim_debug_misc( __FUNCTION__, "ACL: presence\n" );
      if ( rbud->acl.subscriptions )
        gaim_debug_misc( __FUNCTION__, "ACL: subscriptions\n" );
      if ( rbud->acl.subscribe_others )
        gaim_debug_misc( __FUNCTION__, "ACL: subscribe-others\n" );
    }
  }
  gaim_debug_misc( __FUNCTION__, "-------------------\n" );
}

/*
 * Free up the RVP portions of a buddy structure. Probably incomplete.
 */
static void rvp_free_buddy( GaimBuddy *buddy ) {
  RVPBuddy *rb;

  if ( buddy == NULL ) {
    /* wtf? */
    return;
  }

  if (( rb = buddy->proto_data ) == NULL ) {
    return;
  }

  if ( rb->principal != NULL ) {
    g_free( rb->principal );
  }

  if ( rb->cookies != NULL ) {
    g_hash_table_destroy( rb->cookies );
  }

  g_free( rb );
  buddy->proto_data = NULL; /* avoid nasty accidents */
}

/*
 * keep track of conversations with people who aren't on the buddy list
 */
static GaimBuddy *rvp_get_temp_buddy1( GaimConnection *gc,
                                       GaimBuddy *buddy,
                                       gboolean createp ) {
  RVPData *rd = gc->proto_data;
  RVPBuddy *rb = buddy->proto_data;
  GaimBuddy *retval = NULL;

  if ( rd->nonbuddy == NULL ) {
    gaim_debug_misc( __FUNCTION__, "creating hash table\n" );
    rd->nonbuddy = g_hash_table_new( g_str_hash, g_str_equal );
  }

  /* we may have to populate the nick field if I've been lazy */
  if ( buddy->name == NULL ) {
    gchar *nick = rindex( rb->principal, '/' );
    if ( nick != NULL ) {
      buddy->name = g_strdup( &nick[1] );
    }
  }

  /* these are edge cases which I should really catch elsewhere */
  if ( rb->buddy == NULL ) {
    rb->buddy = buddy;
  }

  if ( buddy->proto_data == NULL ) {
    buddy->proto_data = rb;
  }

  /* special case: me */
  if ( !strcmp( rb->principal, rd->me.principal )) {
    return rd->me.buddy;
  }

  retval = (GaimBuddy *)g_hash_table_lookup( rd->nonbuddy, rb->principal );
  if ( retval == NULL ) {
    if ( createp == TRUE ) {
      GaimBuddy *newbuddy;

#ifdef DUPBUDDY
      newbuddy = g_new0( GaimBuddy, 1 );
      memcpy( newbuddy, buddy, sizeof( GaimBuddy ));
      newbuddy->proto_data = g_new( RVPBuddy, 1 );
      memcpy( newbuddy->proto_data, buddy->proto_data, sizeof( RVPBuddy ));
#else
      newbuddy = buddy;
#endif

      gaim_debug_misc( __FUNCTION__, "didn't find %s, creating at %p\n",
                       rb->principal, newbuddy );
      g_hash_table_insert( rd->nonbuddy, g_strdup( rb->principal ), newbuddy );
      retval = newbuddy;
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "found %s in hash at %p\n", rb->principal,
                     retval );
  }

  return retval;
}

static GaimBuddy *rvp_get_temp_buddy_create( GaimConnection *gc,
                                             GaimBuddy *buddy ) {
  GaimBuddy *retval = rvp_get_temp_buddy1( gc, buddy, TRUE );

  if ( retval != NULL ) { /* and it shouldn't be */
  } else {
    gaim_debug_error( __FUNCTION__, "Doh. NULL buddy created?\n" );
  }

  return retval;
}

static GaimBuddy *rvp_get_temp_buddy( GaimConnection *gc, GaimBuddy *buddy ) {
  return rvp_get_temp_buddy1( gc, buddy, FALSE );
}

/*
 * parse an XML fragment identifying a contact or contacts
 */
static GaimBuddy **parse_contact( xmlDocPtr doc, xmlNodePtr from ) {
  xmlNodePtr contact = from;
  GaimBuddy **blist = NULL;
  gint bcount = 0;

  while( contact ) {
    xmlNodePtr ptr = contact->xmlChildrenNode;
    GaimBuddy *buddy = NULL;

    while( ptr ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        xmlChar *href = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        if ( href != NULL ) {
          gchar **split;
          gint elt = 0;

          if ( buddy == NULL ) {
            buddy = g_new0( GaimBuddy, 1 );
          }
          buddy->proto_data = g_new0( RVPBuddy, 1 );
          ((RVPBuddy *)buddy->proto_data)->buddy = buddy;
          ((RVPBuddy *)buddy->proto_data)->principal =
            g_strdup( (gchar *)href );

          /* xxx functionise! */
          split = g_strsplit( (gchar *)href, "/", 0 );
          while( split[elt] != NULL ) { elt++; }
          if ( strstr( (gchar *)href, "/instmsg/local" ) != NULL ) {
            /* physical URL */
            if ( elt == 9 ) {
              buddy->name = g_strconcat( split[8], "@", split[2], NULL );
            } else {
              gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                                href );
            }
          } else {
            if ( elt == 6 ) {
              buddy->name = g_strconcat( split[5], "@", split[2], NULL );
            } else {
              gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                                href );
            }
          }

          if ( split ) {
            g_strfreev( split );
          }
          xmlFree( href );
        }
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "description")) {
        xmlChar *desc = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        if ( desc != NULL ) {
          if ( buddy == NULL ) {
            buddy = g_new0( GaimBuddy, 1 );
          }
          buddy->server_alias = g_strdup((gchar *)desc );
          xmlFree( desc );
        }
      }

      ptr = ptr->next;
    }

    blist = g_realloc( blist, sizeof( GaimBuddy *) * ( bcount + 1 ));
    blist[bcount] = buddy;
    bcount++;

    contact = contact->next;
  }

  /* null terminate */
  blist = g_realloc( blist, sizeof( GaimBuddy *) * ( bcount + 1 ));
  blist[bcount] = NULL;

  if ( bcount > 1 ) {
    gaim_debug_misc( __FUNCTION__, "multiple contacts found!\n" );
  }

  return blist;
}

/*
 * build an xml fragment representing a contact
 */
static xmlNodePtr build_contact( gchar *host, gchar *nick, gchar *desc,
                                 gchar *principal ) {
  xmlNodePtr contact, node, content;
  gchar *url;

  if ( principal == NULL ) {
    url = g_strdup_printf( "http://%s/instmsg/aliases/%s", host, nick );
  } else {
    url = g_strdup( principal );
  }
  contact = xmlNewNode( NULL, BAD_CAST "r:contact" );
  node = xmlNewNode( NULL, BAD_CAST "d:href" );
  content = xmlNewText( BAD_CAST url );
  xmlAddChild( node, content );
  xmlAddChild( contact, node );

  node = xmlNewNode( NULL, BAD_CAST "r:description" );
  if ( desc != NULL ) {
    content = xmlNewText( BAD_CAST desc );
    xmlAddChild( node, content );
  }
  xmlAddChild( contact, node );

  return contact;
}

/*
 * look up a buddy by rvp name, principal, or session-id
 */
static GaimBuddy *rvp_find_buddy( GaimConnection *gc, gchar *name,
                                  gchar *principal, gchar *session ) {
  GaimBuddyList *blist;
  GaimBlistNode *group, *cnode, *bnode;
  GaimBuddy *retval = NULL;

#if 0
  RVPData *rd = gc->proto_data;

  /* this'd be nice, but currently breaks some assumptions elsewhere */
  if (( name != NULL && !strcmp( name, rd->me.buddy->name )) ||
      ( principal != NULL && !strcmp( principal, rd->me.principal ))) {
    return rd->me.buddy;
  }
#endif

  blist = gaim_get_blist();
  for ( group = blist->root; group; group = group->next ) {
    for ( cnode = group->child; cnode; cnode = cnode->next ) {
      if ( !GAIM_BLIST_NODE_IS_CONTACT(cnode)) {
        continue;
      }
      for ( bnode = cnode->child; bnode; bnode = bnode->next ) {
        RVPBuddy *target;

        if (!GAIM_BLIST_NODE_IS_BUDDY( bnode )) {
          continue;
        }
        retval = (GaimBuddy *)bnode;

        if ( retval->account != gaim_connection_get_account( gc )) {
          continue;
        }

        if ( name != NULL ) {
          if ( !strcmp( retval->name, name )) {
            goto gotcha;
          } else {
            continue;
          }
        }

        /*
         * Anything below here requires that the RVPBuddy structure
         * is valid.
         */
        target = retval->proto_data;

        /*
         * async notifications can cause this to be triggered
         * before we've set up the proto_data for the buddies
         */
        if ( target == NULL ) {
          continue;
        }

        if ( principal != NULL ) {
          if ( target->principal != NULL &&
               !strcmp( target->principal, principal )) {
            goto gotcha;
          }
        }

        if ( session != NULL ) {
          if ( target->sessionid != NULL &&
               !strcmp( target->sessionid, session )) {
            goto gotcha;
          }
        }
      }
    }
  }

  /* avoid nasty accidents */
  retval = NULL;

 gotcha:
  return retval;
}

#define rvp_find_buddy_by_name( g, n ) rvp_find_buddy( g, n, NULL, NULL )
#define rvp_find_buddy_by_principal( g, p ) rvp_find_buddy( g, NULL, p, NULL )
#define rvp_find_buddy_by_session( g, s ) rvp_find_buddy( g, NULL, NULL, s )

/*
 * parse the head of an XML response. ROOT is the expected root node.
 */
static xmlDocPtr parse_preamble( const gchar *buffer, gint size,
                                 const gchar *root, xmlNodePtr *cur ) {
  xmlDocPtr doc = NULL;

#ifdef LOUD
  gaim_debug_misc( __FUNCTION__, "\n%*.*s\n", size, size, buffer );
#endif

  /* by preference, use xmlReadMemory as it can be told not to whine
     about the invalid 'DAV:' namespace */
#ifdef HAVE_XMLREADMEMORY
  if (( doc = xmlReadMemory( buffer, size, NULL, NULL, XML_PARSE_NOWARNING )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "Document not parsed successfully.\n" );
    return NULL;
  }
#else
  if (( doc = xmlParseMemory( buffer, size )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "Document not parsed successfully.\n" );
    return NULL;
  }
#endif

  if (( *cur = xmlDocGetRootElement( doc )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "empty document\n" );
    xmlFreeDoc( doc );
    return NULL;
  }

  if ( xmlStrcmp( (*cur)->name, BAD_CAST root )) {
    gaim_debug_error( __FUNCTION__, "unexpected doc root\n" );
    xmlFreeDoc( doc );
    return NULL;
  }

  return doc;
}

/*
 * Parse a SUBSCRIPTIONS response to get our buddy list
 */
static void parse_subscriptions( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc;
  xmlNodePtr cur, subscription;

  gaim_debug_misc( __FUNCTION__, "enter\n" );

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "subscriptions", &cur )) == NULL ) {
    goto out;
  }

  if (( subscription = cur->xmlChildrenNode ) == NULL ) {
    gaim_debug_misc( __FUNCTION__, "no subscriptions\n" );
    goto out;
  }

  while( subscription != NULL ) {
    xmlNodePtr ptr = subscription->xmlChildrenNode;
    gint subs_id, timeout;
    gchar *principal = NULL, *href = NULL;

#ifdef DEBUG
    xmlDebugDumpNode( stderr, subscription, TRUE );
#endif

    while( ptr ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "subscription-id" )) {
        xmlChar *sid = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        subs_id = atoi((gchar *)sid );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        href = (gchar *)xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "timeout" )) {
        xmlChar *to = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        timeout = atoi((gchar *)to );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "principal" )) {
        xmlNodePtr rvpp = ptr->xmlChildrenNode;
        if ( rvpp != NULL && !xmlStrcmp( rvpp->name,
                                         BAD_CAST "rvp-principal" )) {
          principal =
            (gchar *)xmlNodeListGetString( doc, rvpp->xmlChildrenNode, 1 );
        }
      } else {
        /* unknown item, ignore for now FIXME */
        gaim_debug_error( __FUNCTION__, "what's a %s?\n", (gchar *)ptr->name );
      }

      ptr = ptr->next;
    }

    /* this is the only thing we really need to retrieve the
       subscription */
    if ( principal != NULL ) {
      gchar **split = NULL;
      gint elt = 0;
      GaimBuddy *buddy;
      gboolean temp = FALSE;

      /* hmm. this doesn't currently recognise ME... */
      if (( buddy = rvp_find_buddy_by_principal( gc, principal )) == NULL ) {
        buddy = g_new0( GaimBuddy, 1 );
        temp = TRUE;
      }

      if ( buddy->proto_data == NULL ) {
        buddy->proto_data = g_new0( RVPBuddy, 1 );
        ((RVPBuddy *)buddy->proto_data)->buddy = buddy;
      }

      ((RVPBuddy *)buddy->proto_data)->principal = g_strdup( principal );

      /* funnily enough, doing this causes all manner of
         trouble. don't do it.
      ((RVPBuddy *)buddy->proto_data)->subs_id = subs_id;
      */

      if ( temp ) {
        buddy = rvp_get_temp_buddy_create( gc, buddy );
      }

      /* From Microsoft's RVP doc:

      There are two types of URLs: logical URLs and physical URLs. A
      logical URL is the default and determines which domain the node
      is located on. For example, an e-mail address such as
      beverlyj@domain1.com could map to a logical URL
      http://im.domain1.com/instmsg/aliases/beverlyj.

      A physical URL can be the same as a logical URL, or it can
      provide extra information as to which home server is responsible
      for the entity. A physical URL is used when a router redirects
      requests for a specific entity. For example, if the e-mail
      address garrettv@domain2.com is within a domain with a router
      and several home servers, any requests to the router can be
      redirected to the physical URL
      http://imhome1.domain2.com/instmsg/local/im.domain2.com/instmsg/aliases/garrettv.
      This physical URL is then used to access the node at the home
      server imhome1.domain2.com.
      */
      split = g_strsplit( principal, "/", 0 );
      while( split[elt] != NULL ) { elt++; }
      if ( strstr( principal, "/instmsg/local" ) != NULL ) {
        /* physical URL */
        if ( elt == 9 ) {
          buddy->name = g_strconcat( split[8], "@", split[2], NULL );
        } else {
          gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                            principal );
        }
      } else {
        if ( elt == 6 ) {
          buddy->name = g_strconcat( split[5], "@", split[2], NULL );
        } else {
          gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                            principal );
        }
      }

      if ( split ) {
        g_strfreev( split );
      }
    } else {
      gaim_debug_error( __FUNCTION__, "principal null\n" );
    }

    subscription = subscription->next;
  }

 out:
  if ( doc != NULL ) {
    xmlFree( doc );
  }

  gaim_debug_misc( __FUNCTION__, "out\n" );
}

/*
 * Parse an ACL collection
 */
static void parse_acls( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  RVPData *rd = gc->proto_data;
  xmlDocPtr doc;
  xmlNodePtr cur, acl, ace;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len, "rvpacl",
                              &cur )) == NULL ) {
    return;
  }

  if (( acl = cur->xmlChildrenNode ) == NULL ) {
    gaim_debug_error( __FUNCTION__, "no acls\n" );
    goto out;
  }

  while( acl ) {
    ace = acl->xmlChildrenNode;

    while( ace ) {
      GaimBuddy *buddy = NULL;
      xmlChar *rvp_principal = NULL;
      xmlNodePtr credentials = NULL;
      xmlNodePtr bits = ace->xmlChildrenNode;

      /* skip the 'inheritance: none' crap */
      if ( xmlStrcmp( ace->name, BAD_CAST "ace" )) {
        gaim_debug_misc( __FUNCTION__, "skipping %s\n", (gchar *)ace->name );
        goto loop;
      }

      /* the horror, the horror */
      if ( !xmlStrcmp( bits->name, BAD_CAST "principal" )) {
        xmlNodePtr pbits = bits->xmlChildrenNode;
        while( pbits != NULL ) {
          if ( !xmlStrcmp( pbits->name, BAD_CAST "rvp-principal" )) {
            if (( rvp_principal =
                  xmlNodeListGetString( doc, pbits->xmlChildrenNode, 1 ))
                != NULL ) {
              if ( !xmlStrcmp( BAD_CAST rd->me.principal, rvp_principal )) {
                buddy = rd->me.buddy;
              } else {
                buddy = rvp_find_buddy_by_principal( gc,
                                                     (gchar *)rvp_principal );

                if ( buddy == NULL ) {
                  /* need to fake enough of a GaimBuddy structure to
                     keep this happy, which maybe I should do elsewhere... */
                  GaimBuddy *fake = g_new0( GaimBuddy, 1 );
                  RVPBuddy *faker = g_new0( RVPBuddy, 1 );
                  fake->proto_data = faker;
                  faker->principal = (gchar *)rvp_principal;
                  buddy = rvp_get_temp_buddy_create( gc, fake );
                }
              }
            }
          } else if ( !xmlStrcmp( pbits->name, BAD_CAST "credentials" )) {
            credentials = pbits->xmlChildrenNode;
          }
          pbits = pbits->next;
        }
      }

      if ( buddy ) {
        RVPBuddy *rbuddy = buddy->proto_data;
        xmlNodePtr grants = NULL, denies = NULL;

        while( bits ) {
          if ( !xmlStrcmp( bits->name, BAD_CAST "grant" )) {
            grants = bits->xmlChildrenNode;
            gaim_debug_misc( __FUNCTION__, "got grant\n" );
          } else if ( !xmlStrcmp( bits->name, BAD_CAST "deny" )) {
            denies = bits->xmlChildrenNode;
            gaim_debug_misc( __FUNCTION__, "got deny\n" );
          }
          bits = bits->next;
        }

        while ( credentials ) {
          if ( !xmlStrcmp( credentials->name, BAD_CAST "assertion" )) {
            rbuddy->acl.credentials.assertion = TRUE;
          } else if ( !xmlStrcmp( credentials->name, BAD_CAST "digest" )) {
            rbuddy->acl.credentials.digest = TRUE;
          } else if ( !xmlStrcmp( credentials->name, BAD_CAST "ntlm" )) {
            rbuddy->acl.credentials.ntlm = TRUE;
          }

          credentials = credentials->next;
        }

        while ( grants ) {
          if ( !xmlStrcmp( grants->name, BAD_CAST "list" )) {
            rbuddy->acl.list = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "read" )) {
            rbuddy->acl.read = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "write" )) {
            rbuddy->acl.write = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "send-to" )) {
            rbuddy->acl.send_to = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "receive-from" )) {
            rbuddy->acl.receive_from = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "readacl" )) {
            rbuddy->acl.readacl = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "writeacl" )) {
            rbuddy->acl.writeacl = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "presence" )) {
            rbuddy->acl.presence = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "subscriptions" )) {
            rbuddy->acl.subscriptions = TRUE;
          } else if ( !xmlStrcmp( grants->name,
                                  BAD_CAST "subscribe-others" )) {
            rbuddy->acl.subscribe_others = TRUE;
          } else {
            gaim_debug_warning( __FUNCTION__, "unrecognised ACL %s\n",
                                grants->name );
          }

          grants = grants->next;
        }

        while ( denies ) {
          if ( !xmlStrcmp( denies->name, BAD_CAST "list" )) {
            rbuddy->acl.list = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "read" )) {
            rbuddy->acl.read = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "write" )) {
            rbuddy->acl.write = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "send-to" )) {
            rbuddy->acl.send_to = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "receive-from" )) {
            rbuddy->acl.receive_from = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "readacl" )) {
            rbuddy->acl.readacl = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "writeacl" )) {
            rbuddy->acl.writeacl = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "presence" )) {
            rbuddy->acl.presence = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "subscriptions" )) {
            rbuddy->acl.subscriptions = FALSE;
          } else if ( !xmlStrcmp( denies->name,
                                  BAD_CAST "subscribe-others" )) {
            rbuddy->acl.subscribe_others = FALSE;
          } else {
            gaim_debug_warning( __FUNCTION__, "unrecognised ACL %s\n",
                                denies->name );
          }

          denies = denies->next;
        }

#ifdef DEBUG
        rvp_dump_buddy( buddy );
#endif

      } else {
        if ( rvp_principal != NULL ) {
          /* shouldn't ever see this as we're faking up a buddy struct
             above */
          gaim_debug_misc( __FUNCTION__, "can't find buddy for %s\n",
                           (gchar *)rvp_principal );
#ifdef DEBUG
          xmlDebugDumpNode( stderr, ace, TRUE );
#endif
        } else {
          /* HMM. Maybe I should create a fake buddy for this as it'd
             give me plenty space to keep the ACLs */
          gaim_debug_misc( __FUNCTION__,
                           "Null Principal, probably generic acl\n" );
#ifdef DEBUG
          xmlDebugDumpNode( stderr, ace, TRUE );
#endif
        }
      }

    loop:
      ace = ace->next;
    }

    acl = acl->next;
  }

 out:
  xmlFreeDoc( doc );
}

/*
 * Parse a 207 Multistatus response and apply its data to whatever
 * we've got in core.
 */
static void parse_multistatus( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc;
  xmlNodePtr cur, response;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "multistatus", &cur )) == NULL ) {
    return;
  }

  /* children of a multistatus:
   * - response {set}
   *   - href of principal concerned
   *   - propstat
   *     - prop
   *       - r:state, d:displayname
   *     - status
   */
  response = cur->xmlChildrenNode;

  while ( response != NULL ) {
    xmlNodePtr ptr, href = NULL, propstat = NULL, prop = NULL, status = NULL;
    xmlChar *val;
    RVPData *rvpdata = gc->proto_data;
    RVPBuddy *target = NULL, *metoo = NULL;
    GaimBuddy *buddy = NULL;

    if ( xmlStrcmp( response->name, BAD_CAST "response" )) {
      gaim_debug_error( __FUNCTION__, "expected <response> here\n" );
      goto loop;
    }

    ptr = response->xmlChildrenNode;
    while ( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        href = ptr;
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "propstat" )) {
        propstat = ptr;
      } else {
        gaim_debug_error( __FUNCTION__, "unknown response part\n" );
      }

      ptr = ptr->next;
    }

    if ( href == NULL || propstat == NULL ) {
      goto loop;
    }

    if (( val =
          xmlNodeListGetString( doc, href->xmlChildrenNode, 1 )) == NULL ) {
      goto loop;
    }

    /*
     * using metoo here in case you've got yourself on your buddy
     * list. This has triggered a bug in the past, btw, and I don't
     * recommend it.
     */
    target = NULL;
    metoo = NULL;

    if ( !strcmp( rvpdata->me.principal, (char *)val )) {
      metoo = &(rvpdata->me);
    }

    buddy = rvp_find_buddy_by_principal( gc, (char *)val );
    if ( buddy == NULL ) {
      target = metoo;
    } else {
      target = buddy->proto_data;
    }

    if ( target == NULL ) {
      gaim_debug_warning( __FUNCTION__, "don't know this principal (%s)\n",
                          (char *)val );
      goto loop;
    } else if ( target == metoo ) {
      gaim_debug_misc( __FUNCTION__, "setting state for myself\n" );
    }

    /* now handle the props */

    /* first, get the status */
    ptr = propstat->xmlChildrenNode;
    while( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "prop" )) {
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "status" )) {
        status = ptr;
      } else {
        gaim_debug_error( __FUNCTION__, "unknown response part\n" );
      }

      ptr = ptr->next;
    }

    if ( status == NULL ) {
      gaim_debug_error( __FUNCTION__, "No status in property\n" );
      goto loop;
    }

    /* xxx extract status code */

    /* now do the props */
    ptr = propstat->xmlChildrenNode;
    while( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "prop" )) {
        prop = ptr->xmlChildrenNode;

        while( prop != NULL ) {
          if ( !xmlStrcmp( prop->name, BAD_CAST "displayname" )) {
            if (( val =
                  xmlNodeListGetString( doc, prop->xmlChildrenNode, 1 ))
                != NULL ) {
              if ( target->buddy->server_alias != NULL ) {
                g_free( target->buddy->server_alias );
              }
              target->buddy->server_alias = g_strdup( (gchar *)val );

              if ( metoo != NULL && target != metoo ) {
                if ( metoo->buddy->server_alias != NULL ) {
                  g_free( metoo->buddy->server_alias );
                }
                metoo->buddy->server_alias = g_strdup( (gchar *)val );
                gaim_account_set_alias( gaim_connection_get_account( gc ),
                                        g_strdup(( gchar *) val ));
              }

              xmlFree( val );
            }

            /* update the buddy list view */
            if ( target != metoo && target->buddy->server_alias != NULL ) {
              gaim_blist_update_buddy_icon( target->buddy );
            }

          } else if ( !xmlStrcmp( prop->name, BAD_CAST "state" )) {
            /* big state or little state? */
            xmlNodePtr state = NULL, snode = prop->xmlChildrenNode;

            while( snode ) {
              if ( !xmlStrcmp( snode->name, BAD_CAST "leased-value" )) {
                /* big state */
                xmlNodePtr bits = snode->xmlChildrenNode;

                while( bits ) {
                  if ( !xmlStrcmp( bits->name, BAD_CAST "value" )) {
                    state = bits->xmlChildrenNode;
                  } else if ( !xmlStrcmp( bits->name, BAD_CAST "timeout" )) {
                    xmlChar *to =
                      xmlNodeListGetString( doc, bits->xmlChildrenNode , 1 );
                    if ( to != NULL ) {
                      rvpdata->view_expiry = atol( (gchar *)to ) +
                        time( NULL );
                      gaim_debug_misc( __FUNCTION__,
                                       "view expires in %u seconds (%s)\n",
                                       atol((gchar *)to ), to );
                    }
                  }

                  bits = bits->next;
                }
              } else if ( !xmlStrcmp( snode->name, BAD_CAST "view-id" )) {
                gchar *viewid =
                  (gchar *)xmlNodeListGetString( doc, snode->xmlChildrenNode,
                                                 1 );
                if ( viewid != NULL ) {
                  rvpdata->view_id = atol( viewid );
                  gaim_debug_misc( __FUNCTION__, "my view id: %d\n",
                                   rvpdata->view_id );
                } else {
                  gaim_debug_error( __FUNCTION__, "can't parse view-id\n" );
                }
              } else {
                /* little state */
                state = snode;
              }

              snode = snode->next;
            }

            if ( state != NULL && target != metoo ) {
              rvp_set_buddy_status( gc, target->buddy->name,
                                    (char *)state->name );
            }
          } else {
            xmlChar *v = xmlNodeListGetString( doc, prop->xmlChildrenNode, 1 );

            if ( !xmlStrcmp( prop->name, BAD_CAST "email" )) {
              target->email = g_strdup( (gchar *)v );
            } else if ( !xmlStrcmp( prop->name, BAD_CAST "mobile-state" )) {
              target->mobile_state = atoi((gchar *)v );
            } else if ( !xmlStrcmp( prop->name,
                                    BAD_CAST "mobile-description" )) {
              target->mobile_description = g_strdup((gchar *)v);
            } else {
              gaim_debug_warning( __FUNCTION__, "unknown property %s = %s\n",
                                  (gchar *)prop->name, (gchar *)v );
            }
          }
          prop = prop->next;
        }
      }

      ptr = ptr->next;
    }

  loop:
    response = response->next;
  }

  xmlFreeDoc( doc );
}

/*
 * extract a given header from the header block and return its content
 *
 * This is a bit wasteful in that it reparses the headers every time
 * it's called, but hey. Also, it expects both inputs to be null-terminated.
 */
static gchar *get_header_content( gchar *hblock, gchar *header ) {
  gchar **headers = NULL;
  gchar *retval = NULL, *headercolon = NULL;

  headercolon = g_strconcat( header, ":", NULL );

  /* fast return */
  if ( !strstr( hblock, headercolon )) {
    g_free( headercolon );
    return NULL;
  }

  /* technically \r\n, but be lenient in what you accept &c. */
  headers = g_strsplit( hblock, "\n", 0 );

  if ( headers != NULL ) {
    gint i;
    for ( i = 0; headers[ i ]; i++ ) {
      /* fixme hardcoded == teh wrongness */
      /* the correct way to handle this is to either fold the headers
         together, or to allow for a param specifying which instance
         of a header I'm interested in */
      if ( !strncmp( headers[i], "WWW-Authenticate: Negotiate",
                     strlen( "WWW-Authenticate: Negotiate" ))) {
        continue;
      }

      if ( !strncmp( headers[ i ], headercolon, strlen( headercolon ))) {
        retval = g_strdup( &headers[i][strlen( headercolon )]);
        /* xxx glib doc doesn't make it clear if this is done in-place */
        retval = g_strstrip( retval );

        break;
      }
    }

    g_strfreev( headers );
  }
  g_free( headercolon );

  return retval;
}

/*
 * build a principal from an email address. don't forget to free it
 * when you're done.
 */
static gchar *rvp_principal_from_address( GaimConnection *gc,
                                          const gchar *addr ) {
  gchar **bits = g_strsplit( addr, "@", 2 );
  gchar *srvname, *host = NULL, *principal = NULL;
  srvrec *service = NULL;
  gint port = RVP_PORT;

  if ( bits == NULL || bits[0] == NULL ) {
    goto out;
  }

  if ( bits[1] == NULL ) {
    const gchar *newname = rvp_normalize( gaim_connection_get_account( gc ),
                                          addr );
    g_strfreev( bits );
    bits = g_strsplit( newname, "@", 2 );
  }

  host = g_strdup( bits[1] );

  srvname = g_strconcat( "_rvp._tcp.", bits[1], NULL );
  service = gethostbysrv( srvname, NULL );
  g_free( srvname );

  if ( service->host == NULL ) {
    struct hostent *h;
    if (( h = gethostbyname( bits[1] )) == NULL ) {
      goto out;
    }
  } else {
    g_free( host );
    host = g_strdup( service->host );
    port = service->port;
  }

  if ( port == RVP_PORT ) {
    principal = g_strdup_printf( "http://%s/instmsg/aliases/%s",
                                 host, bits[0] );
  } else {
    principal = g_strdup_printf( "http://%s:%d/instmsg/aliases/%s",
                                 host, port, bits[0] );
  }


 out:
  if ( host ) {
    g_free( host );
  }
  if ( bits != NULL ) {
    g_strfreev( bits );
  }

  return principal;
}

/*
 * Build a minimal buddy structure from an email address
 */
static GaimBuddy *rvp_buddy_from_address( GaimConnection *gc,
                                          const gchar *fullname ) {
  GaimBuddy *retval = NULL;
  RVPBuddy *rb = NULL;
  gchar **bits = g_strsplit( fullname, "@", 2 );
  gchar *principal = NULL;
  RVPData *rd;

  gaim_debug_misc( __FUNCTION__, "creating buddy from address %s\n",
                   fullname );

  principal = rvp_principal_from_address( gc, fullname );

  if ( gc ) {
    rd = gc->proto_data;
    retval = (GaimBuddy *)g_hash_table_lookup( rd->nonbuddy, principal );
  } else {
    gaim_debug_error( __FUNCTION__, "called with no gc. wtf?\n" );
  }

  if ( retval == NULL ) {
    retval = g_new0( GaimBuddy, 1 );
    retval->proto_data = g_new0( RVPBuddy, 1 );
    rb = retval->proto_data;
    rb->principal = principal;
    if ( gc ) {
      retval->name =
        g_strdup( rvp_normalize( gaim_connection_get_account( gc ),
                                 fullname ));
    } else {
      retval->name = g_strdup( bits[0] ); /* XXX */
      gaim_debug_error( __FUNCTION__, "called with no gc. 2wtf?\n" );
    }
  } else {
    g_free( principal );
  }

  if ( bits != NULL ) {
    g_strfreev( bits );
  }
  return retval;
}

/*
 * file receive stuff
 */
static void rvp_xfer_init_recv( GaimXfer *xfer ) {
  RVPInvite *inv = xfer->data;
  GaimAccount *ac = gaim_xfer_get_account( xfer );
  GaimConnection *gc = gaim_account_get_connection( ac );

  gaim_debug_misc( __FUNCTION__, "Enter\n" );
  /* we need an IP address from the server */
  if ( gaim_xfer_get_remote_ip( xfer ) == NULL ) {
    gchar *filexfer;
    filexfer = g_strdup_printf( "Invitation-Command: ACCEPT\r\n"
                                "Invitation-Cookie: %d\r\n"
                                "Launch-Application: FALSE\r\n"
                                "Request-Data: IP-Address:\r\n\r\n",
                                inv->cookie );
    rvp_send_notify( gc, inv->who, RVP_MSG_INVITE, filexfer, 0 );
    g_free( filexfer );
  }
}

static void rvp_xfer_end( GaimXfer *xfer ) {
  gaim_debug_misc( __FUNCTION__, "Enter\n" );
}

static void rvp_xfer_cancel_recv( GaimXfer *xfer ) {
  RVPInvite *inv = xfer->data;
  GaimAccount *ac = gaim_xfer_get_account( xfer );
  GaimConnection *gc = gaim_account_get_connection( ac );
  gchar *filexfer;

  gaim_debug_misc( __FUNCTION__, "Enter\n" );

  /* fixme: check what the reason is */
  /* REJECT => user rejected
     FTTIMEOUT => timeout in file transfer
     TIMEOUT => user retracted ?
  */

  filexfer = g_strdup_printf( "Invitation-Command: CANCEL\r\n"
                              "Invitation-Cookie: %d\r\n"
                              "Cancel-Code: REJECT\r\n\r\n",
                              inv->cookie );
  rvp_send_notify( gc, inv->who, RVP_MSG_INVITE, filexfer, 0 );
  g_free( filexfer );
}

#if GAIM_MAJOR_VERSION < 2
static void rvp_xfer_ack_recv( GaimXfer *xfer, const char *buffer,
                               size_t size ) {
  gaim_debug_misc( __FUNCTION__, "Enter\n" );
}
#else
static void rvp_xfer_ack_recv( GaimXfer *xfer, const guchar *buf,
                               size_t size ) {
  gaim_debug_misc( __FUNCTION__, "Enter\n" );
}
#endif

static void rvp_xfer_cb( gpointer data, int source, GaimInputCondition cond ) {
  GaimXfer *xfer = data;
  RVPInvite *inv = xfer->data;
  char buf;
  gboolean eof = FALSE;

  if ( read( source, &buf, 1 ) > 0 || errno == EWOULDBLOCK ) {
    if ( errno == EWOULDBLOCK ) {
      errno = 0;
      return;
    }
    fprintf( stderr, "XFER-> %c\n", buf );
  } else if ( errno != ETIMEDOUT ) {
    eof = TRUE;
  } else {
    gaim_input_remove( inv->xferinp );
    close( inv->xfersock );
  }

  if ( eof ) {
    gaim_input_remove( inv->xferinp );
    close( inv->xfersock );
  }
}

static void rvp_send_callback( gpointer data, gint source,
                               GaimInputCondition cond ) {
  GaimXfer *xfer = data;
  RVPInvite *inv = xfer->data;

  if (( inv->xfersock = accept( source, 0, 0 )) < 0 ) {
    perror( "Accept failed" );
    return;
  }

  inv->xferinp = gaim_input_add( inv->xfersock, GAIM_INPUT_READ, rvp_xfer_cb,
                                 xfer );
}

/*
 * parse an invite message
 */
static void rvp_parse_invite( GaimConnection *gc, GaimBuddy *buddy,
                              gchar *data ) {
  RVPBuddy *rb;
  RVPData *rd = gc->proto_data;
  gchar **lines = g_strsplit( data, "\n", 0 );
  gint i;
  GHashTable *headers = g_hash_table_new( g_str_hash, g_str_equal );
  enum rvp_invite_type inv = RVP_INV_UNKNOWN;
  RVPInvite *invite;
  gchar *guid;
  gint cookie;

  if ( buddy != NULL ) {
    rb = buddy->proto_data;
  } else {
    gaim_debug_error( __FUNCTION__, "Woah, no buddy\n" );
    goto out;
  }

  /* this code should really be split out into
     "make_hash_from_headers" or some such */
  for ( i = 0; lines[ i ]; i++ ) {
    gchar **bits = g_strsplit( lines[ i ], ": ", 2 );
    if ( bits == NULL || bits[ 0 ] == NULL || bits[ 1 ] == NULL ) {
      gaim_debug_error( __FUNCTION__, "Unparsable line %s\n", lines[ i ] );
      goto endloop;
    }

    if ( !strcmp( bits[ 0 ], "Invitation-Command" )) {
      if ( !strcmp( bits[ 1 ], "INVITE" )) {
        inv = RVP_INV_INVITE;
      } else if ( !strcmp( bits[ 1 ], "ACCEPT" )) {
        inv = RVP_INV_ACCEPT;
      } else if ( !strcmp( bits[ 1 ], "CANCEL" )) {
        inv = RVP_INV_CANCEL;
      } else {
        gaim_debug_error( __FUNCTION__, "Unrecognised invite command '%s'\n",
                          bits[ 1 ] );
      }
      /* we won't be using this header any further */
      goto endloop;
    } else if ( !strcmp( bits[ 0 ], "Invitation-Cookie" )) {
      cookie = atoi( bits[ 1 ] ); /* xxx make sure this is parseable! */
      goto endloop;
    } else {
      gaim_debug_warning( __FUNCTION__, "accepting verbatim '%s'\n",
                          lines[ i ] );
    }

    g_hash_table_insert( headers, g_strdup( bits[ 0 ] ),
                         g_strdup( bits[ 1 ] ));

  endloop:
    g_strfreev( bits );
  }

  /* so, what have we got? */
  switch( inv ) {
  case RVP_INV_INVITE:
    /* what kind of invite? */
    guid = g_hash_table_lookup( headers, "Application-GUID" );
    if ( strcmp( guid, RVP_GUID_FILE_TRANSFER )) {
      /* flat-out reject these. fixme, see if there's an unknown guid
         cancel-code */
      gchar *msg = g_strdup_printf( "Invitation-Command: CANCEL\r\n"
                                    "Invitation-Cookie: %d\r\n"
                                    "Cancel-Code: REJECT\r\n\r\n",
                                    cookie );
      gaim_debug_misc( __FUNCTION__, "Unsupported GUID\n" );
      rvp_send_notify( gc, buddy->name, RVP_MSG_INVITE, msg, 0 );
    } else {
      invite = g_new0( RVPInvite, 1 );
      invite->cookie = cookie;
      invite->who = buddy->name;

      if ( !strcmp( guid, RVP_GUID_FILE_TRANSFER )) {
        GaimXfer *xfer;
        gchar *filename = g_hash_table_lookup( headers, "Application-File" );
        gchar *filesize =
          g_hash_table_lookup( headers, "Application-FileSize" );

        xfer = gaim_xfer_new( gaim_connection_get_account( gc ),
                              GAIM_XFER_RECEIVE,
                              buddy->server_alias );
        gaim_xfer_set_filename( xfer, filename );
        gaim_xfer_set_size( xfer, atoi( filesize ));
        gaim_xfer_set_init_fnc( xfer, rvp_xfer_init_recv );
        gaim_xfer_set_end_fnc( xfer, rvp_xfer_end );
        gaim_xfer_set_request_denied_fnc( xfer, rvp_xfer_cancel_recv );
        gaim_xfer_set_cancel_recv_fnc( xfer, rvp_xfer_cancel_recv );
        gaim_xfer_set_ack_fnc( xfer, rvp_xfer_ack_recv );

        /* keep a handle on it */
        xfer->data = invite;
        invite->data = xfer;

        gaim_xfer_request( xfer );
      }

      /* stash the cookie */
      if ( rb->cookies == NULL ) {
        rb->cookies = g_hash_table_new( g_direct_hash, g_direct_equal );
      }
      g_hash_table_insert( rb->cookies, (gpointer)cookie, invite );
    }
    break;

  case RVP_INV_ACCEPT:
  case RVP_INV_CANCEL:
    if (( rb->cookies == NULL ) ||
        (( invite = g_hash_table_lookup( rb->cookies,
                                         (gconstpointer)cookie )) == NULL )) {
      gaim_debug_error( __FUNCTION__,
                        "got an invite/accept for an unknown cookie\n" );
    } else {
      GaimXfer *xfer = invite->data;

      if ( inv == RVP_INV_CANCEL ) {
        if ( xfer != NULL &&
             gaim_xfer_get_status( xfer ) != GAIM_XFER_STATUS_CANCEL_LOCAL &&
             gaim_xfer_get_status( xfer ) != GAIM_XFER_STATUS_CANCEL_REMOTE ) {
          gaim_xfer_cancel_remote( xfer );
        }
        g_hash_table_remove( rb->cookies, (gconstpointer)cookie );
      } else if ( inv == RVP_INV_ACCEPT ) {
        gchar *ipaddr = g_hash_table_lookup( headers, "IP-Address" );
        gchar *port = g_hash_table_lookup( headers, "Port" );
        gchar *portx = g_hash_table_lookup( headers, "PortX" );
        gchar *auth = g_hash_table_lookup( headers, "AuthCookie" );
        gchar *req = g_hash_table_lookup( headers, "Request-Data" );

        if ( xfer != NULL && gaim_xfer_get_type( xfer ) == GAIM_XFER_SEND ) {
          if ( req != NULL && !strcmp( req, "IP-Address:" )) {
            gint port, portx;
            gchar *msg;
            struct sockaddr_in name, namex;

            /* set up listeners */
            invite->authcookie = random_integer( 1, 1 << 31 );
            invite->sock = socket( AF_INET, SOCK_STREAM, 0 );
            invite->sockx = socket( AF_INET, SOCK_STREAM, 0 );
            fcntl( invite->sock, F_SETFL, O_NONBLOCK );
            fcntl( invite->sockx, F_SETFL, O_NONBLOCK );

            bzero( &name, sizeof( name ));
            name.sin_family = AF_INET;
            name.sin_port = 0;
            name.sin_addr.s_addr = INADDR_ANY;
            memcpy( &namex, &name, sizeof( name ));

            port = rvp_bind_local_socket( gc, invite->sock, &name, 0 );
            portx = rvp_bind_local_socket( gc, invite->sockx, &namex, 0 );

            listen( invite->sock, 5 );
            listen( invite->sockx, 5 );

            invite->inp = gaim_input_add( invite->sock, GAIM_INPUT_READ,
                                          rvp_send_callback, xfer );
            invite->inpx = gaim_input_add( invite->sockx, GAIM_INPUT_READ,
                                           rvp_send_callback, xfer );

            /* strange things happen with the presence of PortX, so
               I'm leaving it out until I've figured out what's broken */
            msg = g_strdup_printf( "Invitation-Command: ACCEPT\r\n"
                                   "Invitation-Cookie: %d\r\n"
                                   "IP-Address: %s\r\n"
                                   "Port: %d\r\n"
                                   /*"PortX: %d\r\n"*/
                                   "AuthCookie: %d\r\n"
                                   "Launch-Application: FALSE\r\n"
                                   "Request-Data: IP-Address:\r\n\r\n",
                                   cookie, rd->client_host,
                                   port, /*portx,*/ invite->authcookie );

            rvp_send_notify( gc, invite->who, RVP_MSG_INVITE, msg, 0 );
          }
        }

        if ( ipaddr == NULL || port == NULL || portx == NULL ||
             auth == NULL ) {
          gaim_debug_misc( __FUNCTION__, "missing data\n" );
        }
      }
    }
    break;

  default:
    gaim_debug_misc( __FUNCTION__, "Don't know what to do with this\n" );
    break;
  }

 out:
  if ( lines != NULL ) {
    g_strfreev( lines );
  }
}

/*
 * parse a notification
 */
static gint parse_notify( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc = NULL;
  xmlNodePtr cur, notification;
  gint retval = 200;
  GaimBuddy *buddy = NULL, *from_contact = NULL, **from_contacts = NULL,
    *to_contact = NULL, **to_contacts = NULL;
  RVPData *rd = gc->proto_data;
  GaimConversation *conv = NULL;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "notification", &cur )) == NULL ) {
    retval = 400;
    goto out;
  }

  notification = cur->xmlChildrenNode;

  while( notification != NULL ) {
    /* what type of notification? */
    if ( !xmlStrcmp( notification->name, BAD_CAST "message" ) ||
         !xmlStrcmp( notification->name, BAD_CAST "propnotification" )) {
      xmlNodePtr ptr = notification->xmlChildrenNode;
      xmlNodePtr from = NULL, to = NULL, msg = NULL, props = NULL;
      xmlChar *val = NULL;

      gaim_debug_misc( __FUNCTION__, "Notify type: %s\n",
                       (gchar *)notification->name );

      while( ptr ) {
        if ( !xmlStrcmp( ptr->name, BAD_CAST "notification-from" )) {
          from = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name,
                                BAD_CAST "notification-to" )) {
          to = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name, BAD_CAST "msgbody" )) {
          msg = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name, BAD_CAST "propertyupdate" )) {
          props = ptr->xmlChildrenNode;
        } else {
          gaim_debug_error( __FUNCTION__, "unknown notification part\n" );
        }
        ptr = ptr->next;
      }

      /* nothing to do? bail! */
      if ( !msg && !props ) {
        goto loop;
      }

      if ( !from ) { /* go digging for a session id! */
        goto loop;
      } else {
        /* who the hell are you? */
        from_contacts = parse_contact( doc, from );
        if ( from_contacts == NULL ) {
          /* xxx */
          gaim_debug_error( __FUNCTION__, "no/unparseable sender\n" );
          goto loop;
        } else {
          /* we do not expect to see multiple senders... */
          RVPBuddy *rb_from;
          from_contact = from_contacts[ 0 ];
          rb_from = from_contact->proto_data;

          buddy = rvp_find_buddy_by_principal( gc, rb_from->principal );
        }
      }

      if ( to ) {
        to_contacts = parse_contact( doc, to );
        if ( to_contacts == NULL ) {
          gaim_debug_error( __FUNCTION__, "no/unparseable recipient\n" );
        } else {
          to_contact = to_contacts[ 0 ];
        }
      }

      if ( msg != NULL && !xmlStrcmp( msg->name, BAD_CAST "mime-data" )) {
        if (( val =
              xmlNodeListGetString( doc, msg->xmlChildrenNode, 1 )) != NULL ) {
          gchar **parts = NULL;

          if ( strstr( (gchar *)val, "\r\n\r\n" )) {
            parts = g_strsplit( (gchar *)val, "\r\n\r\n", 2 );
          } else if ( strstr( (gchar *)val, "\n\n" )) {
            parts = g_strsplit( (gchar *)val, "\n\n", 2 );
          } else {
            /* what sort of mutant are you, exactly? */
            gaim_debug_error( __FUNCTION__,
                              "can't find header delimiter in \n%s\n",
                              (gchar *)val );
          }

          if ( parts[0] != NULL ) {
            gchar *typist = NULL;
            gchar *sessid = NULL;
            gchar *format = NULL;
            gint id = 0; /* for chats */
            gboolean is_chat = FALSE;
            gint msgtype = RVP_MSG_UNKNOWN;
            gchar *contenttype;

            contenttype = get_header_content( parts[0], "Content-Type" );
            if ( contenttype ) {
              if ( !strncmp( contenttype, "text/plain",
                             strlen( "text/plain" ))) {
                /* the rest may or may not specify a charset */
                /* fixme should grab it if it's there and behave
                   accordingly */
                msgtype = RVP_MSG_IM;
              } else if ( !strncmp( contenttype, "text/x-msmsgscontrol",
                                    strlen( "text/x-msmsgscontrol" ))) {
                msgtype = RVP_MSG_TYPING;
              } else if ( !strncmp( contenttype, "text/x-msmsgsinvite",
                                    strlen( "text/x-msmsgsinvite" ))) {
                msgtype = RVP_MSG_INVITE;
              } else if ( !strncmp( contenttype, "text/x-imleave",
                                    strlen( "text/x-imleave" ))) {
                msgtype = RVP_MSG_CHAT_LEAVE;
              }
            }

            /* todo: MIME-Version (ignored) */
            sessid = get_header_content( parts[0], "Session-Id" );
            typist = get_header_content( parts[0], "TypingUser" );
            format = get_header_content( parts[0], "X-MMS-IM-Format" );

            if ( sessid != NULL ) {
              /* get rid of the curlies */
              if ( sessid[0] == '{' ) {
                sessid = memmove( sessid, &sessid[1], strlen( sessid ) - 1 );
              }
              if ( sessid[strlen( sessid ) - 2] == '}' ) {
                sessid[strlen( sessid ) - 2] = '\0';
              }

              gaim_debug_misc( __FUNCTION__, "session id %s\n", sessid );

              /* see if this is a multi-user chat! */
              if ( rd->chats ) {
                conv = g_hash_table_lookup( rd->chats, sessid );

                if ( conv != NULL ) {
                  if ( gaim_conversation_get_type( conv ) ==
                       GAIM_CONV_TYPE_CHAT ) {
                    gaim_debug_misc( __FUNCTION__,
                                     "that's apparently a chat\n" );
                    id = gaim_conv_chat_get_id( GAIM_CONV_CHAT( conv ));
                    is_chat = TRUE;
                  } else {

                  }
                }
              }

              /* you've been press-ganged into a multi-user chat */
              if ( !is_chat && to_contacts[0] != NULL &&
                   to_contacts[1] != NULL && msgtype != RVP_MSG_CHAT_LEAVE ) {
                GHashTable *chat = g_hash_table_new( g_str_hash, g_str_equal );
                is_chat = TRUE;

                g_hash_table_insert( chat, "sessid", sessid );
                g_hash_table_insert( chat, "from", from_contact );
                g_hash_table_insert( chat, "others", to_contacts );

                serv_join_chat( gc, chat );
              }

              /* see if we've got a handle to work with */
              if ( from_contact ) {
                RVPBuddy *p;
                if ( buddy == NULL ) {
                  buddy = rvp_get_temp_buddy_create( gc, from_contact );
                }

                /* this won't help for users not in your buddy list
                   because gaim explictly ignores their
                   server_aliases. OH well. */
                if ( buddy->server_alias == NULL ) {
                  buddy->server_alias = g_strdup( from_contact->server_alias );
                }

                p = buddy->proto_data;

                if ( p ) {
                  /* trash the old session ID */
                  if ( p->sessionid != NULL ) {
                    g_free( p->sessionid );
                  }
                  p->sessionid = g_strdup( sessid );
                } else {
                  /* can't happen, ish */
                  gaim_debug_error( __FUNCTION__, "no proto data\n" );
                }
              } else {
                /* errrr. */
              }
            }

            if ( parts[1] != NULL ) {
              gchar *msgbody;

              switch( msgtype ) {
              case RVP_MSG_IM:
#ifdef FONT_NOODLING
                /* silly formatting dance */
                if ( format != NULL ) {
                  gchar *open = NULL, *close = NULL;
                  gchar **bits;

                  bits = g_strsplit( format, "; ", 0 );

                  open = g_strdup( "<FONT " );
                  close = g_strdup( "</FONT>" );

                  if ( bits != NULL ) {
                    int b;

                    for ( b = 0; bits[ b ] != NULL; b++ ) {
                      if ( !strncmp( bits[ b ], "FN=", 3 )) {
                        int len, c;
                        gchar *newopen;

                        newopen = g_strconcat( open, "FACE='", NULL );
                        g_free( open );

                        len = strlen( newopen );
                        newopen = g_realloc( newopen, len +
                                             strlen( bits[b] - 2 ));
                        for ( c = 3; c < strlen( bits[b] ); c++ ) {
                          if ( !strncmp( &bits[b][c], "%20", 3 )) {
                            newopen[len] = ' ';
                            c += 2;
                          } else {
                            newopen[len] = bits[b][c];
                          }
                          len++;
                        }
                        newopen[len] = '\0';
                        open = g_strconcat( newopen, "' ", NULL );
                      } else if ( !strncmp( bits[ b ], "CO=", 3 )) {
                        /* blue/green/red, in hex, with leading zeros
                           omitted */
                        int len, clen;
                        gchar *newopen;

                        newopen = g_strconcat( open, "COLOR='#000000'", NULL );
                        g_free( open );
                        open = newopen;
                        len = strlen( open );
                        clen = strlen( bits[ b ] ) - 3;

                        /* I could do this with a complex for loop,
                           I'm sure. modulus would be involved, and
                           it'd be (more) incomprehensible. */
                        if ( clen > 0 ) { /* have blue component */
                          open[ len - 6 ] = bits[b][clen + 2];
                        }
                        if ( clen > 1 ) {
                          open[ len - 7 ] = bits[b][clen + 1];
                        }
                        if ( clen > 2 ) { /* have green */
                          open[ len - 4 ] = bits[b][clen];
                        }
                        if ( clen > 3 ) {
                          open[ len - 5 ] = bits[b][clen - 1];
                        }
                        if ( clen > 4 ) { /* have red */
                          open[ len - 2 ] = bits[b][clen - 2];
                        }
                        if ( clen > 5 ) {
                          open[ len - 3 ] = bits[b][clen - 3];
                        }
                      }
                    }

                    g_strfreev( bits );
                  }

                  gaim_debug_misc( __FUNCTION__, "formatting: %s\n", open );

                  msgbody = g_strconcat( open, ">", (gchar *)parts[1], close,
                                         NULL );
                  g_free( open );
                  g_free( close );
                } else {
#endif
                  msgbody = (gchar *)parts[1];
#ifdef FONT_NOODLING
                }
#endif

                if ( is_chat ) {
                  serv_got_chat_in( gc, id, buddy->name, 0, msgbody,
                                    time( NULL ));
                } else {
                  serv_got_im( gc, buddy->name, msgbody, 0, time( NULL ));
                  /*
                  GaimConversation *conv;
                  conv = gaim_find_conversation_with_account( buddy->name, gaim_connection_get_account( gc ));
                  g_hash_table_replace( rd->chats, sessid, conv );
                  */
                }
                /* fixme auto reply if away? */
#ifdef FONT_NOODLING
                if ( format != NULL ) {
                  g_free( msgbody );
                }
#endif
                break;

              case RVP_MSG_TYPING:
                /* hrm. not only is there an at-sign in the typist, but
                   the domain is technically incorrect. I could probably
                   look it up in the buddy list or via the session ID,
                   but for now: */
                if ( typist == NULL ) {
                  /* this happens when a remote user pops up a dialog
                     box to talk to you but hasn't started typing
                     yet. this can be used to pop a dialog box
                     asking if you want to receive an unsolicited
                     message. if there are multiple recipients for
                     this then it's an invite to a multi-user chat */
                  if ( to_contacts[ 1 ] != NULL ) {
                    GHashTable *chat = g_hash_table_new( g_str_hash,
                                                         g_str_equal );

                    g_hash_table_insert( chat, "sessid", sessid );
                    g_hash_table_insert( chat, "from", from_contact );
                    g_hash_table_insert( chat, "other", to_contacts );

                    /* the proper windows-like logic here is to only
                       do a serv_join_chat here if we're already in
                       conversation; otherwise we should note the
                       sessid and punt until we receive a message from
                       this conversation */
                    /* serv_got_chat_invite( gc, "Multi-User Conversation",
                       from_contact->name, NULL, chat ); */
                    serv_join_chat( gc, chat );
                  } else {
                    /* nothing to do for now, since ACLs seem to be
                       managed on the server? */
                  }
                } else {
                  if ( buddy != NULL && buddy->name == NULL ) {
                    gaim_debug_misc( __FUNCTION__, "added buddy name %s\n",
                                     typist );
                    buddy->name = g_strdup( typist );
                  }

                  serv_got_typing( gc, typist, EXCHANGE_TYPING_TIMEOUT,
                                   GAIM_TYPING );
                }
                break;

              case RVP_MSG_CHAT_LEAVE:
                /* user has left a chat */
                rvp_chat_user_left( gc, id, from_contact );
                break;

                /* invite gets used for a bunch of messages... */
              case RVP_MSG_INVITE:
                rvp_parse_invite( gc, buddy, (char *)parts[1] );
                break;

              default:
                gaim_debug_misc( __FUNCTION__, "unknown messagetype %s\n",
                                 contenttype );
                break;
              }
            } else {
              gaim_debug_misc( __FUNCTION__, "empty message\n" );
            }

            g_strfreev( parts );
          } else {
            /* no parts, but we've already warned */
          }
          xmlFree( val );
        }
      } else {
        xmlNodePtr prop = props->xmlChildrenNode;
        while( prop ) {
          xmlNodePtr name = prop->xmlChildrenNode;
          xmlNodePtr value = NULL;

          if ( name != NULL ) {
            value = name->xmlChildrenNode;
          } else {
            /* we expect a null here, as it happens */
            val = xmlNodeListGetString( doc, name, 1 );
            if ( val != NULL ) {
              gaim_debug_misc( __FUNCTION__, "non-null prop\n" );
            }
            goto proploop;
          }

          if ( buddy == NULL ) {
            buddy = rvp_get_temp_buddy( gc, from_contact );
          }

          if ( !xmlStrcmp( name->name, BAD_CAST "state")) {
            if ( buddy != NULL ) {
              rvp_set_buddy_status( gc, buddy->name,
                                    (gchar *)value->name );
            } else {
              RVPBuddy *rb_from = from_contact->proto_data;
              gaim_debug_misc( __FUNCTION__,
                               "got state for unknown buddy %s\n",
                               rb_from->principal );
            }
          } else {
            gaim_debug_misc( __FUNCTION__, "unknown property %s\n",
                             name->name );
          }

        proploop:
          prop = prop->next;
        }
      }
    } else {
      fprintf( stderr, "unhandled notification type " );
#ifdef DEBUG
      xmlDebugDumpString( stderr, notification->name );
#endif
      fprintf( stderr, "\n" );
#ifdef DEBUG
      xmlDebugDumpNode( stderr, notification, 0 );
#endif
    }

  loop:

#if 0
    /* This stuff isn't right just yet, and I don't know why not. */
    if ( from_contacts ) {
      int i = 0;
      while(( from_contact = from_contacts[ i ] ) != NULL ) {
        if ( from_contact->proto_data ) {
          rvp_free_buddy( from_contact );
        }
        if ( from_contact->name ) {
          g_free( from_contact->name );
        }
        if ( from_contact->server_alias ) {
          g_free( from_contact->server_alias );
        }
      }
      g_free( from_contacts );
      from_contacts = NULL;
    }

    if ( to_contacts ) {
      int i = 0;
      while(( to_contact = to_contacts[ i ] ) != NULL ) {
        if ( to_contact->proto_data ) {
          rvp_free_buddy( to_contact->proto_data );
        }
        if ( to_contact->name ) {
          g_free( to_contact->name );
        }
        if ( to_contact->server_alias ) {
          g_free( to_contact->server_alias );
        }
        g_free( to_contact );
      }
      g_free( to_contacts );
    }
#endif

    notification = notification->next;
  }

 out:
  if ( doc ) {
    xmlFreeDoc( doc );
  }
  gaim_debug_misc( __FUNCTION__, "exit %d\n", retval );

  return retval;
}

/*
 * Using ACCOUNT, generate a normalised version of the username S
 */
static const char *rvp_normalize( const GaimAccount *account, const char *s ) {
  static char buf[BUF_LEN]; /* gaim requirement, alas */
  gchar **bits;
  gchar *host = NULL;

  g_return_val_if_fail( s != NULL, NULL );

  bits = g_strsplit( s, "@", 2 );

  if ( bits == NULL || bits[0] == NULL ) {
    gaim_debug_error( __FUNCTION__, "failed to split %s\n", s );
    abort();
  }

  if ( bits[1] == NULL ) {
    /* no hostname specified: use the same one as our account */
    GaimConnection *gc = gaim_account_get_connection( account );
    RVPData *rd;
    if ( gc != NULL ) {
      rd = gc->proto_data;
      if ( rd->host != NULL ) {
        host = g_strdup( rd->host );
      } else {
        gaim_debug_error( __FUNCTION__, "doooh 1\n" );
        abort();
      }
    } else {
      /* GaimConnection is unavailable, so we'll need to use
         GaimAccount */
      gchar **bits2 = g_strsplit( gaim_account_get_username( account ),
                                  "@", 2 );
      if ( bits2 == NULL || bits[0] == NULL ) {
        gaim_debug_error( __FUNCTION__, "failed to split %s\n",
                          gaim_account_get_username( account ));
        abort();
      }
      if ( bits2[1] == NULL ) {
        /* you're really starting to annoy me */
        host = g_strdup( gaim_account_get_string( account, "host", NULL ));
      } else {
        host = g_strdup( bits2[1] );
      }
      g_strfreev( bits2 );
    }
  } else {
    host = g_strdup( bits[1] );
  }

  g_snprintf( buf, sizeof( buf ), "%s@%s", bits[0], host );
  g_strfreev( bits );
  g_free( host );

  return buf;
}

/*
 * generate a session ID
 * from the expired RFC:
 *
 * 10.2.19.  Session-ID
 *
 *  Sent with a NOTIFY message when there is no subscription-ID; used
 *  to maintain context in replies to that notification.
 *
 * The Microsoft documentation has no information on this. From
 * observation, Microsoft Messenger preserves session ID only for as
 * long as its login persists. As soon as you sign out and back in
 * again, it creates a new session ID.
 */
static gchar *rvp_get_sessid( void ) {
  gchar *my_sessid = g_malloc0( 37 );
  int i, r;

  for (i = 0; i < 36; i++) {
    if ( i == 8 || i == 13 || i == 18 || i == 23 )
      my_sessid[i] = '-';
    else {
      r = random_integer( 0, 15 );
      my_sessid[ i ] = HEX_ELM[ r ];
    }
  }

  gaim_debug_misc( __FUNCTION__, "generated new ID %s\n", my_sessid );

  return my_sessid;
}

/*
 * Return the name of an icon for this protocol. This is also used for
 * looking up smileys and for logging, so while I'd like to piggyback
 * on MSN, it just confuses things.
 * not really sure why this is a /function/...
 */
static const char *rvp_list_icon( GaimAccount *a, GaimBuddy *b ) {
  return "rvp";
}

/*
 * Figure out what state a buddy is in
 */
static void rvp_list_emblems( GaimBuddy *b, char **se, char **sw, char **nw, char **ne ) {
  int away_type = b->uc;
  char *emblems[4] = { NULL, NULL, NULL, NULL };
  int i = 0;

  /* fixme: align this more closely with the microsoft icons */
  if ( b->present == GAIM_BUDDY_OFFLINE ) {
    emblems[i++] = "offline";
  } else if ( away_type == RVP_BUSY || away_type == RVP_PHONE ) {
    emblems[i++] = "occupied";
  } else if ( away_type != RVP_ONLINE ) {
    emblems[i++] = "away";
  }

  *se = emblems[0];
  *sw = emblems[1];
  *nw = emblems[2];
  *ne = emblems[3];
}

/*
 * fixme: find out when this function is called
 */
static char *rvp_status_text( GaimBuddy *b ) {
  RVPBuddy *rb;

  rb = b->proto_data;
  if ( rb != NULL ) {
    if ( rb->state != NULL ) {
      /* find the current state in our list of states */
      gint i = 0;
      while( awaymsgs[i].tag != NULL ) {
        if ( !strcmp( awaymsgs[i].tag, rb->state )) {
          return g_strdup( awaymsgs[i].text );
        }
        i++;
      }
    }
  }

  /* Fall through */
  return NULL;
}

/*
 * close the RVP connection
 */
static void rvp_close( GaimConnection *gc ) {
  RVPData *rd = gc->proto_data;

  gaim_debug_misc( __FUNCTION__, "enter\n" );

#if 0
  if ( GAIM_CONNECTION_IS_CONNECTED( gc )) {
    /* missing:
       - unsubscribe each user
       - propatch call?
       - cancel any file transfers
    */
    gaim_debug_misc( __FUNCTION__, "sending self unsub\n" );

    /* unsubscribe myself */
    rd->me.unsubbed = FALSE;
    rvp_send_request( gc, "UNSUBSCRIBE", NULL );

    /* we shouldn't really have a gtk function in here, but it's
       unavoidable since Gaim doesn't seem to provide a hook to do
       this otherwise. */
    while( !gtk_main_iteration_do( TRUE )) {
      if ( rd->me.unsubbed ) {
        gaim_debug_misc( __FUNCTION__, "unsubbed\n" );
        break;
      }
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "apparently I'm not logged in\n" );
  }
#endif

  /* shut down the listener */
  if ( rd != NULL ) {
    if ( rd->linpa ) {
      gaim_input_remove( rd->linpa );
    }
    rd->linpa = 0;

    if ( rd->listener_fd > 0 ) {
      close( rd->listener_fd );
    }
    rd->listener_fd = 0;

    if ( rd->nonbuddy ) {
      g_hash_table_destroy( rd->nonbuddy );
    }
  }

  gc->proto_data = NULL;

  gaim_debug_misc( __FUNCTION__, "exit\n" );
}

/*
 * Keepalive: repatch view and subscriptions when they're about to
 * expire. Windows client appears to use 5 minutes of slop. Anything
 * less than two from the server makes the Windows client deeply,
 * deeply unhappy.
 */
static void rvp_keepalive( GaimConnection *gc ) {
  struct rvp_data *rd = gc->proto_data;

  if ( rd->view_expiry - time( NULL ) < 120 ) {
    gaim_debug_misc( __FUNCTION__, "view expires in %d seconds, renewing\n",
                     rd->view_expiry - time( NULL ));
    rvp_send_request( gc, "PROPPATCH", NULL );
  }

  /* two minutes. windows client is at least this antsy */
  if ( rd->me.expires - time( NULL ) < 120 ) {
    GaimBuddyList *blist;
    GaimBlistNode *group, *cnode, *bnode;
    GaimBuddy *buddy;

    gaim_debug_misc( __FUNCTION__,
                     "main sub expires in %d seconds, renewing\n",
                     rd->me.expires - time( NULL ));

    /* resubscribe:
       send a SUBSCRIBE for me, and one for everything I'm subscribed
       to, and instead of a Notification-Type header use the existing
       subscription ID. */
    blist = gaim_get_blist();
    for ( group = blist->root; group; group = group->next ) {
      for ( cnode = group->child; cnode; cnode = cnode->next ) {
        if ( !GAIM_BLIST_NODE_IS_CONTACT(cnode)) {
          continue;
        }
        for ( bnode = cnode->child; bnode; bnode = bnode->next ) {
          if (!GAIM_BLIST_NODE_IS_BUDDY( bnode )) {
            continue;
          }

          buddy = (GaimBuddy *)bnode;

          /* only subscribe things attached to my account! */
          if ( buddy->account != gaim_connection_get_account( gc )) {
            continue;
          }

          serv_add_buddy( gc, buddy );
        }
      }
    }
    rvp_send_request( gc, "SUBSCRIBE", NULL );
  }
}

/*
 * ACL routines - currently non-functional xxx
 */
static void rvp_set_permit_deny(GaimConnection *gc) {
  gaim_debug_misc( __FUNCTION__, "RVP: Set permit deny\n");

  /* per server.c:
   * this is called when either you import a buddy list, and make lots
   * of changes that way, or when the user toggles the permit/deny
   * mode in the prefs. In either case you should probably be
   * resetting and resending the permit/deny info when you get this.
   */
}


static void rvp_add_permit(GaimConnection *gc, const char *who) {

}


static void rvp_rem_permit(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "RVP: Rem permit %s\n", who);
}


static void rvp_add_deny(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "RVP: Add deny %s\n", who );
}


static void rvp_rem_deny(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "RVP: Rem deny %s\n", who );
}

/*
 * Populate buddy info window. Doesn't actually *get* information,
 * since that's an asynchronous process and I've not set it up yet.
 *
 * fixme: this should look at the temp buddy data, too
 */
static void rvp_get_info( GaimConnection *gc, const char *who ) {
  GaimBuddy *buddy = rvp_find_buddy_by_name( gc, (gchar *)who );
  RVPBuddy *rb = NULL;
  gchar *info = NULL;
  gchar *title = NULL;

  if ( buddy == NULL ) {
    buddy = rvp_buddy_from_address( gc, who );
  }

  if ( buddy == NULL || buddy->proto_data == NULL ) {
    info = g_strdup_printf( "Sorry, no info available on %s", who );
  } else {
    rb = buddy->proto_data;

    /* I'd like to put an icon here, but I've not read up that yet. */

    info = g_strdup_printf( "Sign-in Name: %s<br>"
                            "Status: %s<br>"
                            "Service: Microsoft Exchange Instant Messaging<br>"
                            "E-mail: %s<br>",
                            buddy->name,
                            rb->state ? rb->state : "unknown",
                            rb->email ? rb->email : "unknown" );

    title = buddy->server_alias ? buddy->server_alias : buddy->name;
  }

  gaim_notify_formatted( gc, _("Buddy Information"), title ? title : "",
                         NULL, info, NULL, NULL );
  g_free( info );
}

/*
 * list of supported away states
 */
static GList *rvp_away_states(GaimConnection *gc) {
  GList *m = NULL;
  gint i = 0;

  /* build the list from our static list */
  while( awaymsgs[ i ].tag != NULL ) {
    m = g_list_append( m, awaymsgs[ i ].text );
    i++;
  }

  return m;
}

/*
 * mark me as being away
 */
static void rvp_set_away( GaimConnection *gc, const char *state,
                          const char *msg ) {
  struct rvp_data *rd = gc->proto_data;
  gchar *away = NULL;
  gchar *args[1];
  gint i = 0;

  if ( gc->away != NULL ) {
    g_free( gc->away );
    gc->away = NULL;
  }

  /* identify the appropriate state tag */
  while( awaymsgs[i].tag != NULL ) {
    if ( !strcmp( awaymsgs[i].text, state )) {
      away = awaymsgs[i].tag;
      break;
    }
    i++;
  }

  /* override: if msg is set, state is away */
  if ( msg ) {
    away = "away";
    gc->away = g_strdup( msg );
  }

  /* fallback: if we're idle, set idle, otherwise set online */
  if ( away == NULL ) {
    if ( gc->is_idle ) {
      away = "idle";
    } else {
      away = "online";
    }
  }

  if ( rd->me.state != NULL ) {
    g_free( rd->me.state );
  }
  rd->me.state = g_strdup( away );

  args[ 0 ] = away;
  rvp_send_request( gc, "PROPPATCH", args );

  return;
}

/*
 * automatically set away when idle
 */
static void rvp_set_idle( GaimConnection *gc, int idle ) {
  RVPData *rd = gc->proto_data;
  char *away;

  if ( rd != NULL && rd->me.state != NULL &&
       strcmp( rd->me.state, "online" ) &&
       strcmp( rd->me.state, "idle" )) {
    /* we're already in some sort of away state */
    gaim_debug_misc( __FUNCTION__, "I am already %s\n", rd->me.state );
    return;
  }

  /* xxx - these should be pulled from awaymsgs */
  if ( idle ) {
    away = "Idle";
  } else {
    away = "Online";
  }

  rvp_set_away( gc, away, NULL );
  rvp_set_buddy_status( gc, rd->me.buddy->name, rd->me.state );
}

/*
 * support functions for context-click
 */
static void rvp_send_email( GaimBlistNode *node, gpointer ignore ) {
  GaimBuddy *buddy = ( GaimBuddy *) node;
  RVPBuddy *rb = buddy->proto_data;
  gchar *command = NULL;

  /* this is all gtknotify does for mailto URLs */
  /* ok, it does more, but it's checking for 50 million mail clients
     and AGAIN they've not exposed the bloody API to us plugin
     authors. Tip: if you write useful code in your app, think about
     whether it might be of use to someone else, too */
  /* even better: gaim_running_gnome actually fails on my
     gnome-running office desktop. Brilliant stuff. */
  if ( gaim_running_gnome() == TRUE ) {
    command = g_strdup_printf( "gnome-open \"%s\"", rb->email );
  } else {
    gaim_notify_error( NULL, _("Error sending mail"),
                       "This only works under Gnome, and even then"
                       "pretty poorly", NULL );
  }
}

/*
 * This is invoked when you context-click on a buddy. Its job is to
 * return menu items for the buddy.
 */
static GList *rvp_buddy_menu( GaimBlistNode *node ) {
  GList *m = NULL;
  GaimBuddy *buddy;
  RVPBuddy *rb;
  GaimBlistNodeAction *act;
  gchar *label;

  if ( !GAIM_BLIST_NODE_IS_BUDDY( node )) {
    return NULL;
  }

  buddy = (GaimBuddy *)node;
  rb = buddy->proto_data;

  /*
   * caution: if you add a bogus user to your list, there's a good
   * chance that there's no valid data in the rb structure.
   */
  if ( rb == NULL ) {
    return NULL;
  }

  if ( rb->email != NULL ) {
    label = g_strdup_printf( "%s (%s)", _("Send E-Mail"), rb->email );
    act = gaim_blist_node_action_new( label, rvp_send_email, NULL );
    m = g_list_append( m, act );
  }

  return m;
}

/*
 * set the state of a buddy
 */
static void rvp_set_buddy_status( GaimConnection *gc, gchar *nick,
                                  char *status ) {
  GaimAccount *ac = gaim_connection_get_account( gc );
  int online, stat;
  GaimBuddy *buddy;
  GaimBuddy *t = NULL;
  RVPBuddy *rb;

  if ( nick == NULL ) { /* wtf? */
    gaim_debug_misc( __FUNCTION__, "ignoring null nick\n" );
    return;
  }

  buddy = gaim_find_buddy( ac, nick );

  if ( buddy == NULL ) {
    /* ain't no buddy of mine! */
    gaim_debug_misc( __FUNCTION__, "non-buddy %s is %s\n", nick, status );
    t = rvp_buddy_from_address( gc, nick );
    buddy = t;
  }

  gaim_debug_misc( __FUNCTION__, "%s is %s\n", buddy->name, status );
  rb = buddy->proto_data;
  if ( rb != NULL ) {

    /* When we've subscribed ourself, it's possible status and
     * rb->state are the same thing.   So, we have to g_strdup()
     * the new status, before we free the old
     */
    gchar *newstate = g_strdup( status );

    if ( rb->state != NULL ) {
      g_free( rb->state );
    }
    rb->state = newstate;
  }

  online = 1;

  /* set buddy online status */
  if (!strcmp(status, "online")) {
    stat = RVP_ONLINE;
  } else if (!strcmp(status, "offline") || !strcmp(status, "invisible")) {
    online = 0;
    stat = RVP_OFFLINE;
  } else if (!strcmp(status, "idle")) {
    stat = RVP_IDLE;
  } else if (!strcmp(status, "away")) {
    stat = RVP_AWAY;
  } else if (!strcmp(status, "busy")) {
    stat = RVP_BUSY;
  } else if (!strcmp(status, "back-soon")) {
    stat = RVP_BRB;
  } else if (!strcmp(status, "on-phone")) {
    stat = RVP_PHONE;
  } else if (!strcmp(status, "at-lunch")) {
    stat = RVP_LUNCH;
  } else {
    /* catch new status and pop up a message window */
    gaim_notify_error(gc, _("Unknown status"), status, NULL );
    gaim_debug_error( __FUNCTION__, "unknown status %s", status );
    stat = RVP_UNKNOWN;
    return;
  }

  serv_got_update(gc, buddy->name, online, 0, 0, 0, stat );

  if ( t != NULL ) {
    gaim_debug_error( __FUNCTION__, "leaking\n" );
  }
}

/*
 * set ACLs for a buddy
 */
static void rvp_set_acl( GaimConnection *gc, RVPBuddy *rb ) {
  xmlDocPtr doc;
  xmlNodePtr root, acl, ace, node1, node2, node3;
  xmlChar *content;
  gint content_length;
  gchar *args[4];

  doc = xmlNewDoc( BAD_CAST "1.0" );
  root = xmlNewNode( NULL, BAD_CAST "r:rvpacl" );
  xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
            BAD_CAST "r" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
            BAD_CAST "a" );
  xmlDocSetRootElement( doc, root );

  acl = xmlNewNode( NULL, BAD_CAST "a:acl" );
  node1 = xmlNewNode( NULL, BAD_CAST "a:inheritance" );
  node2 = xmlNewText( BAD_CAST "none" );
  xmlAddChild( node1, node2 );
  xmlAddChild( acl, node1 );

  ace = xmlNewNode( NULL, BAD_CAST "a:ace" );

  node1 = xmlNewNode( NULL, BAD_CAST "a:principal" );
  node2 = xmlNewNode( NULL, BAD_CAST "a:rvp-principal" );
  node3 = xmlNewText( BAD_CAST rb->principal );
  xmlAddChild( node2, node3 );
  xmlAddChild( node1, node2 );
  xmlAddChild( ace, node1 );

  node2 = xmlNewNode( NULL, BAD_CAST "a:credentials" );
  if ( rb->acl.credentials.assertion ) {
    node3 = xmlNewNode( NULL, BAD_CAST "a:assertion" );
    xmlAddChild( node2, node3 );
  }
  if ( rb->acl.credentials.digest ) {
    node3 = xmlNewNode( NULL, BAD_CAST "a:digest" );
    xmlAddChild( node2, node3 );
  }
  if ( rb->acl.credentials.ntlm ) {
    node3 = xmlNewNode( NULL, BAD_CAST "a:ntlm" );
    xmlAddChild( node2, node3 );
  }
  xmlAddChild( node1, node2 );
  xmlAddChild( ace, node1 );

  node1 = xmlNewNode( NULL, BAD_CAST "a:grant" );

  if ( rb->acl.list ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:list" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.read ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:read" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.write ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:write" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.send_to ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:send-to" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.receive_from ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:receive-from" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.readacl ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:readacl" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.writeacl ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:writeacl" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.presence ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:presence" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.subscriptions ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:subscriptions" );
    xmlAddChild( node1, node2 );
  }
  if ( rb->acl.subscribe_others ) {
    node2 = xmlNewNode( NULL, BAD_CAST "a:subscribe-others" );
    xmlAddChild( node1, node2 );
  }

  xmlAddChild( ace, node1 );

  node1 = xmlNewNode( NULL, BAD_CAST "a:deny" );

#if WAIDER_DOES_BLOCKING
  node2 = xmlNewNode( NULL, BAD_CAST "a:list" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:read" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:write" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:send-to" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:receive-from" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:readacl" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:writeacl" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:presence" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:subscriptions" );
  xmlAddChild( node1, node2 );
  node2 = xmlNewNode( NULL, BAD_CAST "a:subscribe-others" );
  xmlAddChild( node1, node2 );
#endif

  xmlAddChild( ace, node1 );
  xmlAddChild( acl, ace );
  xmlAddChild( root, acl );

  xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
  xmlFreeDoc( doc );

  args[0] = g_strdup_printf( rb->principal );
  args[1] = (gchar *)content;
  args[2] = (gchar *)content_length;
  args[3] = NULL;

  rvp_send_request( gc, "ACL", args );
}

/*
 * add a buddy
 */
static void rvp_add_buddy( GaimConnection *gc, GaimBuddy *buddy,
                           GaimGroup *group ) {
  RVPData *rd = gc->proto_data;
  GaimBuddy *dup;
  RVPBuddy *rb;
  gchar *args[4];

  /* ideally I'd do a duplicate check here, but since I can't do
     anything useful even if I do find one (the buddy will already be
     on the buddy list by the time this code is executed) I've left it
     out. */

  rb = buddy->proto_data;

  if ( rb == NULL ) {
    rb = g_new0( RVPBuddy, 1 );
    rb->buddy = buddy;
    buddy->proto_data = rb;
  }

  if ( rb->principal == NULL ) {
    rb->principal = rvp_principal_from_address( gc, buddy->name );
  }

  dup = rvp_get_temp_buddy( gc, buddy );
  if ( dup ) {
    gaim_debug_misc( __FUNCTION__, "duplicate\n" );
    rvp_dump_buddy( dup );
    /* copy what we've got into the 'real' buddy structure, and delete
       the temp buddy */
    g_hash_table_remove( rd->nonbuddy, rb->principal );
  }

#ifdef DEBUG
  rvp_dump_buddy( buddy );
#endif

  /* don't subscribe if we're not connected (duh) */
  if ( GAIM_CONNECTION_IS_CONNECTED( gc )) {
    xmlDocPtr doc;
    xmlNodePtr root, acl, ace, node1, node2, node3;
    xmlChar *content;
    gint content_length;

    args[0] = (gchar *)buddy;
    args[1] = NULL;
    rvp_send_request( gc, "SUBSCRIBE", args );

    /* We only need to send an ACL if we don't already have an
       appropriate one set. send-to and presence are sufficient for a
       buddy, plus at least one of assertion, digest or ntlm */
    /* also, this should really be tacked onto the return from the
       SUBSCRIBE because otherwise it's putting too much crap on the
       wire in one go. */
    if (( rb->acl.credentials.assertion ||
          rb->acl.credentials.digest ||
          rb->acl.credentials.ntlm ) &&
        rb->acl.send_to == TRUE && rb->acl.presence == TRUE ) {
      return;
    }

    gaim_debug_misc( __FUNCTION__, "user has insufficient privs, ACLing\n" );

    doc = xmlNewDoc( BAD_CAST "1.0" );
    root = xmlNewNode( NULL, BAD_CAST "r:rvpacl" );
    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    acl = xmlNewNode( NULL, BAD_CAST "a:acl" );
    node1 = xmlNewNode( NULL, BAD_CAST "a:inheritance" );
    node2 = xmlNewText( BAD_CAST "none" );
    xmlAddChild( node1, node2 );
    xmlAddChild( acl, node1 );

    ace = xmlNewNode( NULL, BAD_CAST "a:ace" );

    node1 = xmlNewNode( NULL, BAD_CAST "a:principal" );
    node2 = xmlNewNode( NULL, BAD_CAST "a:rvp-principal" );
    node3 = xmlNewText( BAD_CAST rb->principal );
    xmlAddChild( node2, node3 );
    xmlAddChild( node1, node2 );
    xmlAddChild( ace, node1 );

    node2 = xmlNewNode( NULL, BAD_CAST "a:credentials" );
    node3 = xmlNewNode( NULL, BAD_CAST "a:assertion" );
    xmlAddChild( node2, node3 );
    node3 = xmlNewNode( NULL, BAD_CAST "a:digest" );
    xmlAddChild( node2, node3 );
    node3 = xmlNewNode( NULL, BAD_CAST "a:ntlm" );
    xmlAddChild( node2, node3 );
    xmlAddChild( node1, node2 );
    xmlAddChild( ace, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "a:grant" );

    /* brilliant. if you set an ACL on yourself without the writeacl
       permission, you lose the ability to set ACLs for
       yourself. D'oh.
       Also, I've had reports that this is still broken despite my
       attempt to fix it here. Perhaps rb->principal and
       rd->me.principal don't match sufficiently well? */
    if ( !strcmp( rb->principal, rd->me.principal )) {
      node2 = xmlNewNode( NULL, BAD_CAST "a:list" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:read" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:write" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:send-to" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:receive-from" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:readacl" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:writeacl" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:presence" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:subscriptions" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:subscribe-others" );
      xmlAddChild( node1, node2 );
    } else {
      /* for debugging purposes, let's declare the fact */
      gaim_debug_misc( __FUNCTION__, "not me, restrictive ACL applies\n" );
      node2 = xmlNewNode( NULL, BAD_CAST "a:send-to" );
      xmlAddChild( node1, node2 );
      node2 = xmlNewNode( NULL, BAD_CAST "a:presence" );
      xmlAddChild( node1, node2 );
    }

    xmlAddChild( ace, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "a:deny" );
    xmlAddChild( ace, node1 );
    xmlAddChild( acl, ace );
    xmlAddChild( root, acl );

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    args[0] = g_strdup_printf( rb->principal );
    args[1] = (gchar *)content;
    args[2] = (gchar *)content_length;
    args[3] = NULL;

    rvp_send_request( gc, "ACL", args );

    g_free( args[0] );
    /* g_free( content ); DON'T DO THIS! It's freed later. */
  } else {
    gaim_debug_warning( __FUNCTION__, "not connected\n" );
  }
}

/*
 * remove a buddy from the buddy list
 */
static void rvp_rem_buddy(GaimConnection *gc, GaimBuddy *buddy,
                          GaimGroup *group) {
  gchar *args[1];

  gaim_debug_misc( __FUNCTION__, "rem_buddy name: %s\n", buddy->name );

  args[0] = (gchar *)buddy;

  rvp_send_request( gc, "UNSUBSCRIBE", args );
}

static gint gc_strcmp( gconstpointer a, gconstpointer b ) {
  GaimConvChatBuddy *cb = (GaimConvChatBuddy *)a;
  gaim_debug_misc( __FUNCTION__, "%s <=> %s\n", cb->name, b );
  return strcmp( cb->name, b );
}

/*
 * generic notify function
 */
static int rvp_send_notify( GaimConnection *gc, const char *who, gint msgtype,
                            gchar *message, void *data ) {
  int typing = FALSE;
  GaimConversation *conv = NULL;
  GaimConvImFlags flags;
  GaimAccount *ac = gaim_connection_get_account( gc );
  GList *recipients = NULL, *r;
  GaimBuddy *buddy = NULL;
  RVPBuddy *rb = NULL;
  RVPData *rd = gc->proto_data;
  const gchar *fullname;
  gchar *msg = NULL, *sessid = NULL;
  gchar *args[4];
  xmlDocPtr doc;
  xmlNodePtr root, xmlmsg, node, xmlcontact, body, mime;
  xmlChar *content = NULL;
  gint content_length;
  GList *buddies;
  gchar *format = NULL;

  gaim_debug_misc( __FUNCTION__, "Enter\n" );

  switch( msgtype ) {
  case RVP_MSG_TYPING:
    typing = *(int*)data;
    break;

  case RVP_MSG_IM:
    flags = *(GaimConvImFlags*)data;
    break;

  case RVP_MSG_INVITE:
    break;

  case RVP_MSG_CHAT_INVITE:
  case RVP_MSG_CHAT:
  case RVP_MSG_CHAT_LEAVE:
    conv = (GaimConversation *)data;
    sessid = gaim_conversation_get_data( conv, "sessid" );
    break;

  default:
    gaim_debug_error( __FUNCTION__, "unknown msgtype %d (1)\n", msgtype );
    break;
  }

  if ( msgtype == RVP_MSG_TYPING && typing != TRUE ) {
    return EXCHANGE_TYPING_TIMEOUT;
  }

  if ( sessid == NULL || msgtype == RVP_MSG_CHAT_INVITE ) {
    /* pick up a hostname if we don't have one */
    fullname = rvp_normalize( gaim_connection_get_account( gc ), who );

    /*
     * Is it someone we know about?
     */
    buddy = gaim_find_buddy( ac, who );
    if ( buddy == NULL ) {
      buddy = gaim_find_buddy( ac, fullname );
    }

    if ( buddy == NULL ) {
      GaimBuddy *gb = rvp_buddy_from_address( gc, (gchar *)fullname );
      /* step 2, is it someone we're already conversing with? */
      /* gaim_get_conversations returns a glist of conversations from
         which I should be able to go digging for people */

      if ( gb == NULL ) {
        gaim_debug_misc( __FUNCTION__,
                         "rvp_buddy_from_address returned null for %s\n",
                         fullname );
        return ( msgtype == RVP_MSG_TYPING ? EXCHANGE_TYPING_TIMEOUT : 0 );
      }

      /* shouldn't need this any more... */
      buddy = rvp_get_temp_buddy_create( gc, gb );
    }

    rb = buddy->proto_data;
    if ( rb == NULL ) {
      gaim_debug_misc( __FUNCTION__,
                       "you've not done the proto_data hookup\n" );
      return 0;
    }

    if ( rb->sessionid == NULL ) {
      rb->sessionid = rvp_get_sessid();
    }

    if ( sessid == NULL ) {
      sessid = rb->sessionid;
    }
  }

  /* gussy up the message some. we don't support proper formatting,
     and most definitely don't support putting URLs in as <a href="..."> */
  if ( msgtype == RVP_MSG_IM || msgtype == RVP_MSG_CHAT ) {
    format = g_strdup( "FN=MS%%20Shell%%20Dlg; EF=; CO=0; CS=0; PF=0" );
  }

  switch( msgtype ) {
  case RVP_MSG_TYPING:
    msg =
      g_strdup_printf( "MIME-Version: 1.0\r\n"
                       "Content-Type: text/x-msmsgscontrol; charset=UTF-8\r\n"
                       "TypingUser: %s\r\n"
                       "Session-Id: {%s}\r\n\r\n",
                       rd->me.buddy->name,
                       sessid );
    break;

  case RVP_MSG_CHAT_INVITE:
    msg =
      g_strdup_printf( "MIME-Version: 1.0\r\n"
                       "Content-Type: text/x-msmsgscontrol; charset=UTF-8\r\n"
                       "Session-Id: {%s}\r\n\r\n\r\n",
                       sessid );
    break;

  case RVP_MSG_IM:
  case RVP_MSG_CHAT:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/plain; charset=UTF-8\r\n"
                           "X-MMS-IM-Format: %s\r\n"
                           "Session-Id: {%s}\r\n\r\n%s",
                           format,
                           sessid,
                           message );
    break;

  case RVP_MSG_CHAT_LEAVE:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/x-imleave; charset=UTF-8\r\n"
                           "Session-Id: {%s}\r\n\r\n\r\n",
                           sessid );
    break;

  case RVP_MSG_INVITE:
    msg =
      g_strdup_printf( "MIME-Version: 1.0\r\n"
                       "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n"
                       "Session-Id: {%s}\r\n\r\n"
                       "%s",
                       sessid,
                           message );

  default:
    gaim_debug_error( __FUNCTION__, "unknown msgtype %d (2)\n", msgtype );
    break;
  }

  /* xml content */
  doc = xmlNewDoc( BAD_CAST "1.0" );
  root = xmlNewNode( NULL, BAD_CAST "r:notification" );
  xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
            BAD_CAST "r" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
            BAD_CAST "a" );
  xmlDocSetRootElement( doc, root );

  xmlmsg = xmlNewNode( NULL, BAD_CAST "r:message" );

  /* sender */
  node = xmlNewNode( NULL, BAD_CAST "r:notification-from" );
  xmlcontact = build_contact( NULL, NULL, rd->me.buddy->server_alias,
                              rd->me.principal );
  xmlAddChild( node, xmlcontact );
  xmlAddChild( xmlmsg, node );

  /* recipient */
  node = xmlNewNode( NULL, BAD_CAST "r:notification-to" );
  switch( msgtype ) {
  case RVP_MSG_TYPING:
  case RVP_MSG_IM:
  case RVP_MSG_INVITE:
    xmlcontact = build_contact( NULL, NULL, buddy->server_alias,
                                rb->principal );
    xmlAddChild( node, xmlcontact );
    gaim_debug_misc( __FUNCTION__, "recip: %s\n", rb->principal );
    recipients = g_list_append( recipients, g_strdup( rb->principal ));
    break;

  case RVP_MSG_CHAT_INVITE:
  case RVP_MSG_CHAT:
  case RVP_MSG_CHAT_LEAVE:
    if ( msgtype == RVP_MSG_CHAT_INVITE ) {
      recipients = g_list_append( recipients, g_strdup( rb->principal ));
      xmlcontact = build_contact( NULL, NULL, "", rb->principal );
      xmlAddChild( node, xmlcontact );
    }
    /* who's all up in there? */
    buddies = gaim_conv_chat_get_users( GAIM_CONV_CHAT( conv ));
    while( buddies ) {
      GaimConvChatBuddy *cb = (GaimConvChatBuddy *)buddies->data;
      GaimBuddy *b;

      gaim_debug_misc( __FUNCTION__, "buddy: %s\n", cb->name );

      /* gah. chat only stores the username */
      if (( b = rvp_find_buddy_by_name( gc, cb->name )) == NULL ) {
        b = rvp_buddy_from_address( gc, cb->name ); /* leak */
      }

      rb = b->proto_data;
      /* XXX strcmp */
      if ( !g_list_find_custom( recipients, rb->principal,
                                (GCompareFunc)strcmp )) {
        xmlcontact = build_contact( NULL, NULL, "", rb->principal );
        xmlAddChild( node, xmlcontact );
        if ( msgtype != RVP_MSG_CHAT_INVITE ) {
          recipients = g_list_append( recipients, g_strdup( rb->principal ));
        }
      }
      buddies = buddies->next;
    }
    break;

  default:
    gaim_debug_error( __FUNCTION__, "unknown msgtype %d (1)\n", msgtype );
    break;
  }

  xmlAddChild( xmlmsg, node );

  /* message */
  body = xmlNewNode( NULL, BAD_CAST "r:msgbody" );
  mime = xmlNewNode( NULL, BAD_CAST "r:mime-data" );

  node = xmlNewCDataBlock( doc, BAD_CAST msg, strlen( msg ));
  xmlAddChild( mime, node );
  xmlAddChild( body, mime );
  xmlAddChild( xmlmsg, body );
  xmlAddChild( root, xmlmsg );

  /* render it */
  xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
  xmlFreeDoc( doc );

  /* everyone needs a 303^WNOTIFY */
  for ( r = recipients; r != NULL; r = r->next ) {
    gaim_debug_misc( __FUNCTION__, "notify to %s\n", (gchar *)r->data );

    /* shouldn't happen! */
    if ( r->data == NULL ) {
      continue;
    }

    args[0] = (gchar *)r->data;
    /* this gets freed elsewhere, so we have to send a duplicate */
    args[1] = g_malloc0( content_length );
    memcpy( args[1], content, content_length );
    args[2] = (gchar *)content_length;

    /* oooook. lazy waider. */
    if ( message != NULL ) {
      args[3] = g_strdup( message );
    } else {
      args[3] = NULL;
    }

    rvp_send_request( gc, "NOTIFY", args );
  }
  g_free( content );
  g_list_free( recipients );
  recipients = NULL;

  if ( msgtype == RVP_MSG_TYPING ) {
    return EXCHANGE_TYPING_TIMEOUT;
  } else {
    return 1;
  }
}

/*
 * send a typing notification
 */
static int rvp_send_typing( GaimConnection *gc, const char *who, int typing ) {
  return rvp_send_notify( gc, who, RVP_MSG_TYPING, NULL, &typing );
}

/*
 * send an instant message
 */
static int rvp_send_im( GaimConnection *gc, const char *who,
                        const char *message, GaimConvImFlags flags ) {
  return rvp_send_notify( gc, who, RVP_MSG_IM, (gchar *)message, &flags );
}

#ifdef WAIDER_DOES_FILES_NOW
/*
 * send a file
 */
static void rvp_xfer_init_send( GaimXfer *xfer ) {
  GaimAccount *ac = gaim_xfer_get_account( xfer );
  GaimConnection *gc = gaim_account_get_connection( ac );
  gchar *filexfer;
  gint cookie = 0;
  GaimBuddy *buddy;
  RVPBuddy *rb;
  RVPInvite *inv;

  buddy = rvp_find_buddy_by_name( gc, xfer->who );
  if ( buddy == NULL ) {
    gaim_debug_misc( __FUNCTION__, "Glark" );
    gaim_xfer_cancel_local( xfer );
    return;
  }
  rb = buddy->proto_data;
  if ( rb->cookies == NULL ) {
    rb->cookies = g_hash_table_new( g_direct_hash, g_direct_equal );
  }

  while( cookie == 0 ) {
    cookie = random_integer( 1, 1 << 31 );
    if ( g_hash_table_lookup( rb->cookies, (gconstpointer)cookie )) {
      cookie = 0;
    }
  }

  inv = g_new0( RVPInvite, 1 );
  inv->data = xfer;
  inv->cookie = cookie;
  inv->type = RVP_INV_INVITE; /* not really important */
  inv->who = g_strdup( xfer->who );
  xfer->data = inv;
  g_hash_table_insert( rb->cookies, (gpointer)cookie, inv );

  filexfer = g_strdup_printf( "Application-Name: File Transfer\r\n"
                              "Application-GUID: %s\r\n"
                              "Invitation-Command: INVITE\r\n"
                              "Invitation-Cookie: %d\r\n"
                              "Application-File: %s\r\n"
                              "Application-FileSize: %d\r\n\r\n",
                              RVP_GUID_FILE_TRANSFER,
                              cookie,
                              xfer->filename,
                              xfer->size );

  rvp_send_notify( gc, xfer->who, RVP_MSG_INVITE, filexfer, 0 );
  g_free( filexfer );
}

static void rvp_xfer_cancel_send( GaimXfer *xfer ) {
  RVPInvite *inv = xfer->data;
  GaimAccount *ac = gaim_xfer_get_account( xfer );
  GaimConnection *gc = gaim_account_get_connection( ac );
  gchar *filexfer;

  gaim_debug_misc( __FUNCTION__, "Enter\n" );
  filexfer = g_strdup_printf( "Invitation-Command: CANCEL\r\n"
                              "Invitation-Cookie: %d\r\n"
                              "Cancel-Code: TIMEOUT\r\n\r\n",
                              inv->cookie );
  rvp_send_notify( gc, inv->who, RVP_MSG_INVITE, filexfer, 0 );
  g_free( filexfer );
}

static void rvp_xfer_ack_send( GaimXfer *xfer, const char *buffer, size_t size ) {
  gaim_debug_misc( __FUNCTION__, "Enter\n" );
}

static void rvp_send_file( GaimConnection *gc, const char *who,
                           const char *filename ) {
  GaimAccount *ac = gaim_connection_get_account( gc );
  GaimXfer *xfer = gaim_xfer_new( ac, GAIM_XFER_SEND, who );

  /* fixme: it'd be nice to flip 'who' to server_alias */
  gaim_debug_misc( __FUNCTION__, "send %s to %s\n",
                   filename ? filename : "a file", who );

  gaim_xfer_set_init_fnc( xfer, rvp_xfer_init_send );
  gaim_xfer_set_end_fnc( xfer, rvp_xfer_end );
  gaim_xfer_set_cancel_send_fnc( xfer, rvp_xfer_cancel_send );
  gaim_xfer_set_request_denied_fnc( xfer, rvp_xfer_cancel_send );
  gaim_xfer_set_ack_fnc( xfer, rvp_xfer_ack_send );

  if ( filename ) {
    gaim_xfer_request_accepted( xfer, filename );
  } else {
    gaim_xfer_request( xfer );
  }
}

static gboolean rvp_can_receive_file( GaimConnection *gc, const char *who ) {
  gaim_debug_misc( __FUNCTION__, "can receive from %s\n", who );
  return TRUE;
}
#endif

/*
 * invite someone to a chat
 */
void rvp_chat_invite( GaimConnection *gc, int id, const char *msg,
                      const char *who ) {
  GaimAccount *ac = gaim_connection_get_account( gc );
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;
  const gchar *norm;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  norm = rvp_normalize( ac, who );

  if ( !g_list_find_custom( gaim_conv_chat_get_users( GAIM_CONV_CHAT( b )),
                            norm, gc_strcmp )) {
    gaim_debug_misc( __FUNCTION__, "inviting %s to chat %d\n", norm, id );

    rvp_send_notify( gc, norm, RVP_MSG_CHAT_INVITE, NULL, b );
    gaim_conv_chat_add_user( GAIM_CONV_CHAT( b ), g_strdup( norm ), NULL,
                             GAIM_CBFLAGS_NONE, TRUE );
  } else {
    /* tell the user they're a berk */
  }
}

/*
 * join a chat
 */
void rvp_chat_join( GaimConnection *gc, GHashTable *components ) {
  GaimConversation *conv = NULL;
  GaimBuddy *buddy = NULL;
  GaimBuddy **others = NULL;
  gchar *sessid = NULL;
  gint id = 0;
  RVPData *rd = gc->proto_data;

  gaim_debug_misc( __FUNCTION__, "Enter\n" );

  if ( components != NULL ) {
    sessid = g_hash_table_lookup( components, "sessid" );
    buddy = g_hash_table_lookup( components, "from" );
  }

  /* self-created chat */
  if ( sessid == NULL ) {
    sessid = rvp_get_sessid();
  }
  if ( buddy == NULL ) {
    buddy = rd->me.buddy;
  }

  if ( rd->chats == NULL ) {
    rd->chats = g_hash_table_new( g_str_hash, g_str_equal );
    rd->chatid = 1;
  } else {
    conv = g_hash_table_lookup( rd->chats, sessid );
    if ( conv == NULL ) {
      id = rd->chatid;
    } else {
      id = gaim_conv_chat_get_id( GAIM_CONV_CHAT( conv ));
    }
  }

  if ( conv == NULL ) {
    /* not sure about the third param */
    conv = serv_got_joined_chat( gc, id, sessid );
    gaim_debug_misc( __FUNCTION__, "%s created new chat %p\n", buddy->name,
                     conv );
  }
  g_hash_table_replace( rd->chats, g_strdup( sessid ), conv );
  gaim_conversation_set_data( conv, "sessid", sessid );

  if ( !g_list_find_custom( gaim_conv_chat_get_users( GAIM_CONV_CHAT( conv )),
                            buddy->name, gc_strcmp )) {
    gaim_conv_chat_add_user( GAIM_CONV_CHAT( conv ), g_strdup( buddy->name ),
                             NULL, GAIM_CBFLAGS_NONE, TRUE );
    gaim_debug_misc( __FUNCTION__, "added %s to chat\n", buddy->name );
  }

  others = g_hash_table_lookup( components, "others" );
  if ( others != NULL ) {
    int i;
    GList *ulist;

    ulist = gaim_conv_chat_get_users( GAIM_CONV_CHAT( conv ));

    for ( i = 0; others[ i ] != NULL; i++ ) {
      if ( !g_list_find_custom( ulist, others[i]->name, gc_strcmp )) {
        gaim_conv_chat_add_user( GAIM_CONV_CHAT( conv ),
                                 g_strdup( others[i]->name ), NULL,
                                 GAIM_CBFLAGS_NONE, TRUE );
        gaim_debug_misc( __FUNCTION__, "added %s to chat\n", others[i]->name );
      }
    }
  }
}

void rvp_chat_leave( GaimConnection *gc, int id ) {
  RVPData *rd = gc->proto_data;
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;
  gchar *sessid;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  sessid = gaim_conversation_get_data( b, "sessid" );
  gaim_debug_misc( __FUNCTION__, "leaving conv %p\n", b );

  rvp_send_notify( gc, rd->me.buddy->name, RVP_MSG_CHAT_LEAVE, NULL, b );
  g_hash_table_remove( rd->chats, sessid );
}

void rvp_chat_user_left( GaimConnection *gc, int id, GaimBuddy *buddy ) {
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  gaim_conv_chat_remove_user( GAIM_CONV_CHAT(b), buddy->name, NULL );
}

int rvp_chat_send( GaimConnection *gc, int id, const char *message ) {
  RVPData *rd = gc->proto_data;
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return 0;
  }

  gaim_debug_misc( __FUNCTION__, "sending message to conv %p\n", b );

  return rvp_send_notify( gc, rd->me.buddy->name, RVP_MSG_CHAT,
                          (gchar *)message, b );
}

/*
 * Find and bind a socket to an open local socket within the permissible
 * local port range according to the GAIM prefs, or a preferred port.  A
 * preferred_port value <= 0 will use only configured or system-assigned
 * port numbers.
 *
 * Accepts a pre-opened socket fd, and a pointer to a sockaddr_in which
 * should be filled in except for sin_port.
 *
 * Returns the port number eventually bound, or -1 on failure.  On error,
 * the socket has not been closed.
 */
static int rvp_bind_local_socket( GaimConnection *gc, int sock,
                                  struct sockaddr_in *name,
                                  int preferred_port ) {
  int port, port_start, port_end;
  struct sockaddr addr;
  struct sockaddr_in *ptr;
  socklen_t addrlen;
  gboolean fail;
  RVPData *rd = gc->proto_data;

  if ( preferred_port > 0 ) {
    gaim_debug_misc( __FUNCTION__, "Account configured listen port %d\n",
                     preferred_port);
    port_start = preferred_port;
    port_end = preferred_port;
  } else if ( gaim_prefs_get_bool( "/core/network/ports_range_use" )) {
    gaim_debug_misc( __FUNCTION__,
                     "Globally configured listen port range %d:%d\n",
                     rd->port_low, rd->port_high );
    port_start = rd->port_low;
    port_end = rd->port_high;
  } else {
    port_start = 0;
    port_end = 0;
  }

  for ( port = port_start; port <= port_end; port++ ) {
    if ( port < 1024 && port != 0 ) {
      gaim_debug_warning( __FUNCTION__, "ignoring privileged fixed port %d\n",
                          port );
      port = MIN( port_end, 65535 );
    }
    gaim_debug_misc( __FUNCTION__, "Trying to bind to port %d\n", port );
    name->sin_port = htons( port );
    if ( bind( sock, (struct sockaddr *)name, sizeof( *name )) == 0 ) {
      fail = FALSE;
      break;
    } else {
      gaim_debug_error( __FUNCTION__, "Couldn't bind listening port %d: %s\n",
                        port, strerror( errno ) );
      if ( errno != EADDRINUSE || port == port_end ) {
        return -1;
      }
    }
  }

  /* retrieve the port we've set up on */
  addrlen = sizeof( addr );
  if ( getsockname( sock, &addr, &addrlen ) < 0 ) {
    return -1;
  }

  ptr = (struct sockaddr_in *)&addr;
  port = ntohs( ptr->sin_port );
  gaim_debug_misc( __FUNCTION__, "listener port: %d\n", port );
  return port;
}

/*
 * set up the listener for async notification
 */
void rvp_setup_listener( GaimConnection *gc ) {
  struct rvp_data *rd = gc->proto_data;
  struct sockaddr_in name;
  GaimAccount *account;
  int ld;
  int port;

  /* verify that this error works... it's being called from inside a
     callback, so it may not. */
  if (( ld = socket( AF_INET, SOCK_STREAM, 0)) < 0 ) {
    gaim_connection_error(gc, _("Unable to create listener"));
    gc->wants_to_die = TRUE;
    return;
  }

  fcntl( ld, F_SETFL, O_NONBLOCK );

  name.sin_family = AF_INET;
  name.sin_addr.s_addr = INADDR_ANY;

  account = gaim_connection_get_account( gc );
  port = rvp_bind_local_socket( gc, ld, &name,
                                gaim_account_get_int( account, "port", -1 ));
  if ( port < 0 ) {
    gaim_connection_error( gc, _( "Unable to bind listener" ));
    close( ld );
    gc->wants_to_die = TRUE;
    return;
  }

  if ( listen( ld, 5 ) < 0 ) {
    close( ld );
    gaim_connection_error( gc, _( "Unable to set listener" ));
    gc->wants_to_die = TRUE;
    return;
  }

  rd->listener_fd = ld;
  rd->listener_port = port;
  gaim_debug_misc( __FUNCTION__, "listening on port %d, fd %d\n",
                   rd->listener_port, rd->listener_fd );
}

/*
 * this is largely cribbed from Gaim's util.c
 */
static void
destroy_fetch_url_data(GaimFetchUrlData *gfud) {
  gaim_debug_misc( __FUNCTION__, "Enter %p\n", gfud );
  if (gfud->request.webdata         != NULL) g_free(gfud->request.webdata);
  if (gfud->request.header          != NULL) {
    gaim_debug_misc( __FUNCTION__, "freeing header %p\n", gfud->request.header );
    g_free(gfud->request.header);
  }
  if (gfud->response.webdata         != NULL) g_free(gfud->response.webdata);
  if (gfud->response.header          != NULL) g_free(gfud->response.header);
  if (gfud->url             != NULL) g_free(gfud->url);
  if (gfud->method          != NULL) g_free(gfud->method);
  if (gfud->website.address != NULL) g_free(gfud->website.address);
  if (gfud->website.page    != NULL) g_free(gfud->website.page);
  if (gfud->website.user    != NULL) g_free(gfud->website.user);
  if (gfud->website.passwd  != NULL) g_free(gfud->website.passwd);
  g_free(gfud);
  gaim_debug_misc( __FUNCTION__, "Exit\n" );
}

static gboolean parse_content_len( const char *data, size_t data_len,
                                   size_t *content_len ) {
  const char *p = NULL;
  *content_len = 0;

  /* This is still technically wrong, since headers are
   * case-insensitive [RFC 2616, section 4.2], though this ought to
   * catch the normal case.  Note: data is _not_ nul-terminated.
   */
  if ( data_len > 16 ) {
    p = strncasecmp( data, "Content-Length: ", 16 ) == 0 ? data : NULL;
    if ( !p ) {
      p = g_strstr_len( data, data_len, "\nContent-Length: " );
      if ( p )
        p += 1;
    }
  }

  if ( p != NULL && g_strstr_len( p, data_len - ( p - data ), "\n")) {
    *content_len = atoi( &p[strlen( "Content-Length: " )]);
    return TRUE;
  }

  return FALSE;
}

static void url_fetched_cb( gpointer url_data, gint sock,
                            GaimInputCondition cond ) {
  GaimFetchUrlData *gfud = url_data;
  char data;
  gboolean got_eof = FALSE;

  gfud->sock = sock;
  if (sock == -1) {
    gfud->callback(gfud->user_data, gfud, 0);
    if ( !gfud->preserve ) {
      destroy_fetch_url_data(gfud);
    } else {
      gaim_debug_misc( __FUNCTION__, "preserving gfud %p", gfud );
    }
    return;
  }

 resend:
  if (!gfud->sentreq) {
    char *buf;
    gsize content_length = 0;

    gfud->resend = FALSE;

    /* we need to encode the outgoing buffer in the Microsofty way */
    if ( gfud->request.webdata && !gfud->encoded ) {
      GError *error = NULL;
      gsize read, i;
      gchar *bigbuf = NULL;
      gchar *converted = NULL;

      if ( gfud->request.length ) {
        read = gfud->request.length;
      } else {
        gaim_debug_misc( __FUNCTION__, "no length set, you fule\n" );
        read = strlen( gfud->request.webdata );
        gfud->request.length = read;
      }

      /* we need to convert from UCS-2LE. Unfortunately what we've got
         is more like UCS-1, which doesn't exist. So we fake it. */
      bigbuf = g_malloc0( read * 2 );
      for ( i = 0; i < read; i++ ) {
        bigbuf[ i * 2 ] = gfud->request.webdata[ i ];
      }

      converted = g_convert( bigbuf, gfud->request.length * 2,
                             "UTF-8", "UCS-2LE", &read,
                             &content_length, &error );
      if ( converted == NULL ) {
        gaim_notify_error( NULL, _("Error encoding message"),
                           "Something broke. Bug waider about this.",
                           NULL ); /* xxx */
      } else {
        g_free( gfud->request.webdata );

        /* content_length gets set to " the number of bytes stored in
           the output buffer (not including the terminating nul)." */
        /* so, there's a terminating null character, and
           content_length doesn't mention it. We need to make sure of
           the terminating null on account of the use of
           g_strdup_printf below; ideally I'd get rid of it. */
        gfud->request.webdata = g_realloc( converted, content_length + 1 );

        /* we also need to update this, because we may be looping back
           through this function post-authentication */
        gfud->request.length = content_length;
      }
    } else {
      content_length = gfud->request.length;
    }

    gfud->encoded = TRUE;

    buf = g_strdup_printf( "%s %s%s HTTP/1.1\r\n"
                           "Content-Length: %u\r\n"
                           "RVP-Notifications-Version: 0.2\r\n"
                           "Host: %s\r\n%s\r\n%s" ,
                           gfud->method,
                           (gfud->full ? "" : "/"),
                           (gfud->full ? gfud->url : gfud->website.page),
                           content_length,
                           gfud->website.address,
                           gfud->request.header ? gfud->request.header : "",
                           gfud->request.webdata ?
                           gfud->request.webdata : "" );

    gaim_debug_misc( "gaim_url_fetch", "Requesting %p:\n%s\n", gfud, buf );

    write( sock, buf, strlen( buf ));
    fcntl( sock, F_SETFL, O_NONBLOCK );
    gfud->sentreq = TRUE;
    gfud->inpa = gaim_input_add(sock, GAIM_INPUT_READ,
                                url_fetched_cb, url_data);
    gfud->data_len = 4096;
    gfud->response.webdata = g_malloc(gfud->data_len);

    return;
  }

  if ( read( sock, &data, 1 ) > 0 || errno == EWOULDBLOCK ) {
    if ( errno == EWOULDBLOCK ) {
      errno = 0;
      return;
    }

    gfud->len++;

    if (gfud->len == gfud->data_len + 1) {
      gfud->data_len += (gfud->data_len) / 2;

      gfud->response.webdata =
        g_realloc(gfud->response.webdata, gfud->data_len);
    }

    gfud->response.webdata[gfud->len - 1] = data;

    if (!gfud->startsaving) {
      if (data == '\r')
        return;

      if (data == '\n') {
        if (gfud->newline) {
          size_t content_len;
          gfud->startsaving = TRUE;

#ifdef LOUD
          gaim_debug_misc( __FUNCTION__, "Response headers:\n%*.*s\n",
                           gfud->len, gfud->len, gfud->response.webdata );
#else
          gaim_debug_misc( __FUNCTION__, "got headers\n" );
#endif

          /* See if we can find a content length. */
          gfud->has_explicit_data_len =
            parse_content_len( gfud->response.webdata, gfud->len,
                               &content_len );

          if ( !gfud->has_explicit_data_len ) {
            /* We'll stick with an initial 8192 */
            content_len = 8192;
          }

          /* Out with the old... */
          gfud->response.header = gfud->response.webdata;
          gfud->response.header =
            g_realloc( gfud->response.header, gfud->len + 1 );
          gfud->response.header[ gfud->len ] = 0;
          gfud->len = 0;
          gfud->response.webdata = NULL;

          /* In with the new. */
          gfud->data_len = content_len;
          if ( content_len ) {
            gfud->response.webdata = g_try_malloc( gfud->data_len );
          }
          if ( gfud->response.webdata == NULL ) {
            if ( content_len ) {
              gaim_debug_error( "gaim_url_fetch",
                                "Failed to allocate %u bytes: %s\n",
                                gfud->data_len, strerror( errno ));
            }
            gaim_input_remove( gfud->inpa );
            gfud->callback( gfud->user_data, gfud, 0 );

            if ( gfud->resend ) {
              goto resend;
            }

            gaim_debug_misc( __FUNCTION__, "closing socket for %p\n", gfud );
            close(sock);
            if ( !gfud->preserve ) {
              destroy_fetch_url_data( gfud );
            } else {
      gaim_debug_misc( __FUNCTION__, "preserving gfud %p", gfud );
    }

          }
        } else {
          gfud->newline = TRUE;
        }
        return;
      }

      gfud->newline = FALSE;
    } else if ( gfud->has_explicit_data_len && gfud->len == gfud->data_len ) {
      got_eof = TRUE;
    }
  } else if (errno != ETIMEDOUT) {
    got_eof = TRUE;
  } else {
    gaim_input_remove(gfud->inpa);
    gfud->callback(gfud->user_data, gfud, 0);

    if ( gfud->resend ) {
      goto resend;
    }

    gaim_debug_misc( __FUNCTION__, "closing socket for %p\n", gfud );
    close(sock);

    if ( !gfud->preserve ) {
      destroy_fetch_url_data( gfud );
    } else {
      gaim_debug_misc( __FUNCTION__, "preserving gfud %p", gfud );
    }

  }

  if ( got_eof ) {
    gfud->response.webdata = g_realloc(gfud->response.webdata, gfud->len + 1);
    gfud->response.webdata[gfud->len] = 0;
    gaim_input_remove(gfud->inpa);
    gfud->callback(gfud->user_data, gfud, gfud->len);
    if ( gfud->resend ) {
      goto resend;
    }
    gaim_debug_misc( __FUNCTION__, "closing socket for %p\n", gfud );
    close( sock );

    if ( !gfud->preserve ) {
      destroy_fetch_url_data( gfud );
    } else {
      gaim_debug_misc( __FUNCTION__, "preserving gfud %p", gfud );
    }

  }
}

/* cribbed from LWP::Authen::Digest */
/* CAUTION currently does not handle qop */
#ifdef md5_state_t /* stupid */
static gchar *get_auth_digest( GaimFetchUrlData *gfud, gchar *header,
                               gchar *user, gchar *pass ) {
  gchar **auth_param = NULL;
  gint i = 0;
  gchar *nonce = NULL, *realm = NULL;
  gchar *retval = NULL, hexresp[33];
  md5_state_t md5;
  md5_byte_t seekrit1[16], seekrit2[16], response[16];
  gchar *uri;

  /* slightly naive */
  gaim_debug_misc( __FUNCTION__, "parsing header %s\n", header );
  auth_param = g_strsplit( &header[strlen( "Digest ")], ", ", 0 );

  /* find the nonce */
  for ( i = 0; auth_param[i]; i++ ) {
    gaim_debug_misc( __FUNCTION__, "%d: %s\n", i, auth_param[i] );
    if ( g_str_has_prefix( auth_param[i], "nonce=\"" )) {
      nonce = g_strdup( &auth_param[i][7] );
      nonce[strlen( nonce ) - 1] = '\0';
    } else if ( g_str_has_prefix( auth_param[i], "realm=\"" )) {
      realm = g_strdup( &auth_param[i][7] );
      realm[strlen( realm ) - 1] = '\0';
    }
  }

  if ( nonce == NULL || realm == NULL ) {
    gaim_debug_error( __FUNCTION__,
                      "Can't find one of nonce or realm in header\n" );
    return NULL;
  }

  if ( gfud->website.page[0] != '/' ) {
    uri = g_strconcat( "/", gfud->website.page, NULL );
  } else {
    uri = g_strdup( gfud->website.page );
  }

  md5_init( &md5 );
  md5_append( &md5, (md5_byte_t *)user, strlen( user ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)realm, strlen( realm ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)pass, strlen( pass ));
  md5_finish( &md5, seekrit1 );

  /* digest = ( seekrit1 ) */
  /* digest = ( seekrit1, nonce ) */

  md5_init( &md5 );
  md5_append( &md5, (md5_byte_t *)gfud->method, strlen( gfud->method ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)uri, strlen( uri));
  md5_finish( &md5, seekrit2 );

  /* digest = ( seekrit1, nonce, seekrit2 ) */

  md5_init( &md5 );
  for ( i = 0; i < 16; i++ ) {
    gchar *hex = g_strdup_printf( "%02x", seekrit1[i] );
    md5_append( &md5, (md5_byte_t *)hex, 2 );
    g_free( hex );
  }
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)nonce, strlen( nonce ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  for ( i = 0; i < 16; i++ ) {
    gchar *hex = g_strdup_printf( "%02x", seekrit2[i] );
    md5_append( &md5, (md5_byte_t *)hex, 2 );
    g_free( hex );
  }
  md5_finish( &md5, response );

  for ( i = 0; i < 16; i++ ) {
    sprintf( &hexresp[i * 2], "%02x", response[i] );
  }
  hexresp[32] = '\0';

  retval = g_strdup_printf( "Digest "
                            "username=\"%s\", "
                            "realm=\"%s\", "
                            "algorithm=\"%s\", "
                            "uri=\"%s\", "
                            "nonce=\"%s\", "
                            "response=\"%s\"",
                            user,
                            realm,
                            "MD5",
                            uri,
                            nonce,
                            hexresp );

  if ( auth_param != NULL ) {
    g_strfreev( auth_param );
  }

  return retval;
}
#endif

/*
 * handle data from async notifications
 */
static void rvp_async_data( void *udata, GaimFetchUrlData *gfd, size_t size ) {
  GaimConnection *gc = udata;
  GaimAccount *ac = gaim_connection_get_account( gc );
  RVPData *rd = gc->proto_data;
  size_t i;
  const char *p = NULL;
  gboolean is_xml = FALSE;
  gint code = 0;

  /* Oh KWATZ. */
  gaim_debug_misc( __FUNCTION__, "callback on gfud %p\n", gfd );

  /* to be sure this gets done right... */
  gfd->preserve = FALSE;

  /* Rare, but can happen */
  if ( rd == NULL ) {
    gaim_debug_misc( __FUNCTION__, "Woah, this isn't right!\n" );
    return;
  }

  if ( gfd->response.header == NULL ) {
    gchar *msg =
      g_strdup_printf( _("Server returned empty response to %s request"),
                       gfd->method ? gfd->method : _("unknown"));
    gaim_connection_error( gc, msg );
    g_free( msg );

    if ( gfd->method != NULL ) {
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_debug_warning( __FUNCTION__, "dying...\n" );
        gc->wants_to_die = TRUE;
      }
    } else {
      /* connection still dies when this happens, and I'm not clear why */
      gaim_debug_error( __FUNCTION__, "Someone's probably hacking you...\n" );
    }
    return;
  }

  if ( gfd->method != NULL ) {
    /* find out what the status code is */
    if (( !strncmp( gfd->response.header, "HTTP/1.1 ", 9 )) &&
        strlen( gfd->response.header ) >= 12 &&
        ( code = atoi( &gfd->response.header[9] )) > 99 && code < 1000 ) {
      gaim_debug_misc( __FUNCTION__, "got a %d response to %p %s %s\n", code,
                       gfd, gfd->method, gfd->website.page );
    } else {
      gaim_connection_error( gc, _("Unexpected response from server" ));
      return;
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "async notification\n" );
  }

  if ( gfd->response.header == NULL ) {
      gaim_connection_error( gc, _("No headers in response from server" ));
      return;
  }

  /* redirect */
  if ( code == 302 ) {
    gchar *newurl = get_header_content( gfd->response.header, "Location" );
    if ( newurl != NULL ) {
      gchar *oldhost = NULL;
      gaim_debug_warning( __FUNCTION__, "lightly tested redirection code\n" );
      g_free( gfd->response.header );
      gfd->response.header = NULL;
      g_free( gfd->response.webdata );
      gfd->response.webdata = NULL;
      gfd->len = 0;
      gfd->sentreq = FALSE;
      gfd->startsaving = FALSE;
      gfd->newline = FALSE;
      gfd->resend = TRUE;

      /* In theory, we'd patch the buddy that caused the redirect so
         that future messages go to the correct URL. In practice, this
         appears to be a bad idea, so I've not done it. */
      if ( gfd->website.address ) {
        oldhost = g_strdup( gfd->website.address );
      }

      if ( gfd->url             != NULL) g_free( gfd->url );
      if ( gfd->website.address != NULL) g_free( gfd->website.address );
      if ( gfd->website.page    != NULL) g_free( gfd->website.page );
      if ( gfd->website.user    != NULL) g_free( gfd->website.user );
      if ( gfd->website.passwd  != NULL) g_free( gfd->website.passwd );

      gfd->url = g_strdup( newurl );
      gfd->tried_auth = FALSE; /* hmm */
      gaim_url_parse( gfd->url, &gfd->website.address, &gfd->website.port,
                      &gfd->website.page, &gfd->website.user,
                      &gfd->website.passwd );

      if ( oldhost != NULL ) {
        if ( strcmp( oldhost, gfd->website.address )) {
          /* we need to reconnect */
          int sock;

          gaim_debug_misc( __FUNCTION__, "Reconnection required (%s)\n",
                           gfd->website.address );

          if (( sock = gaim_proxy_connect( ac, gfd->website.address,
                                           gfd->website.port, url_fetched_cb,
                                           gfd )) < 0 ) {
            rvp_async_data( gc, gfd, 0 );
          } else {
            gfd->preserve = TRUE;
            gfd->resend = FALSE; /* non-intuitively, this is correct! */
          }
        }
        g_free( oldhost );
      }

      return;
    }
    /* xxx what's to do here? */
    gaim_connection_error( gc, _( "Got a redirect with no location" ));
    return;
  }

  /* look for an auth failure */
  if ( code == 401 ) {
    gchar *newheaders = NULL;
    gchar *authtype = NULL;

    if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
      gaim_connection_update_progress( gc, _("Authenticating"), rd->login_step,
                                       MAX_LOGIN_STEPS );
    }

    /* headers[0] is the HTTP response line */
    authtype = get_header_content( gfd->response.header, "WWW-Authenticate" );
    if ( authtype == NULL ) {
      gaim_connection_error( gc, _("Auth error, but no auth type requested" ));
      return;
    }

    if ( !strcmp( authtype, "NTLM" )) {
      gchar *ntlm = get_ntlm_msg1( rd->domain, rd->client_host );
      rd->auth_type = RVP_AUTH_NTLM;
      newheaders = g_strdup_printf( "Authorization: %s\r\n", ntlm );
      /* gfd->tried_auth = FALSE; HMMM. Fixme */
    } else if ( g_str_has_prefix( authtype, "NTLM " )) {
      gchar *ntlm;

      if ( gfd->tried_auth ) {
        gaim_debug_warning( __FUNCTION__, "tried authing already!\n" );
        /* prompt for a password! */
        gaim_connection_error( gc, _("Auth failed"));
        ac->password = NULL;
        gc->wants_to_die = TRUE;
        return;
      }

      gfd->tried_auth = TRUE;

      ntlm = get_auth_ntlm( authtype, rd->user, ac->password,
                            rd->client_host, rd->domain,
                            rd->authid );

      newheaders = g_strdup_printf( "Authorization: %s\r\n", ntlm );
#ifdef md5_state_t
    } else if ( g_str_has_prefix( authtype, "Digest" )) {
      gchar *digest = NULL;

      if ( gfd->tried_auth ) {
        /* prompt for a password! */
        gaim_connection_error( gc, _("Auth failed"));
        ac->password = NULL;
        gc->wants_to_die = TRUE;
        return;
      }
      gfd->tried_auth = TRUE;

      digest = get_auth_digest( gfd, authtype, rd->user, ac->password );

      newheaders = g_strdup_printf( "Authorization: %s\r\n", digest );
#endif
    } else {
      gaim_connection_error( gc, _("unrecognised auth type"));
      gaim_debug_error( __FUNCTION__, "unrecognised auth type %s\n",
                        authtype );
      gc->wants_to_die = TRUE;
      return;
    }

    /* a bit silly, esp since newheaders MUST NOT be NULL */
    if ( newheaders != NULL ) {
      gchar **oh = g_strsplit( gfd->request.header, "\r\n", 0 );
      gchar **nh = g_strsplit( newheaders, "\r\n", 0 );
      gchar *newh = NULL, *build = NULL;
      int j;

      if ( nh != NULL && oh != NULL ) {
        for ( j = 0; oh[j] != NULL ; j++ ) {
          gchar **bits = g_strsplit( oh[j], ":", 2 ); /* overkill */
          gchar *hdr = oh[j];

          if ( bits[0] == NULL ) {
            continue;
          }

          for ( i = 0; nh[i]; i++ ) {
            gchar **b2 = g_strsplit( nh[i], ":", 2 );
            if ( b2[0] == NULL ) {
              continue;
            }
            if ( !strcmp( b2[0], bits[0] )) {
              hdr = nh[i];
            }
            g_strfreev( b2 );
          }

          if ( hdr && strlen( hdr )) {
            if ( newh == NULL ) {
              newh = g_strdup_printf( "%s\r\n", hdr );
            } else {
              build = g_strdup_printf( "%s%s\r\n", newh, hdr );
              g_free( newh );
              newh = build;
            }
          }

          g_strfreev( bits );
        }

        for ( i = 0; nh[i]; i++ ) {
          gchar **b2 = g_strsplit( nh[i], ":", 2 );
          gchar *hdr = nh[i];

          for ( j = 0; oh[j]; j++ ) {
            gchar **bits = g_strsplit( oh[j], ":", 2 ); /* overkill */

            if ( bits[0] == NULL ) {
              continue;
            }

            if ( b2[0] == NULL ) {
              continue;
            }
            if ( !strcmp( b2[0], bits[0] )) {
              hdr = NULL;
            }
            g_strfreev( bits );
          }

          if ( hdr && strlen( hdr )) {
            if ( newh == NULL ) {
              newh = g_strdup_printf( "%s\r\n", hdr );
            } else {
              build = g_strdup_printf( "%s%s\r\n", newh, hdr );
              g_free( newh );
              newh = build;
            }
          }

          g_strfreev( b2 );
        }

        g_free( gfd->request.header );
        gfd->request.header = g_strdup( newh );
        g_free( newh );


        g_strfreev( oh );
        g_strfreev( nh );
        g_free( newheaders );
      } else {
        gaim_debug_misc( __FUNCTION__, "split %s failed\n", newheaders );
      }

      gfd->data_len = 0;
      gfd->len = 0;
      gfd->sentreq = FALSE;
      gfd->startsaving = FALSE;
      gfd->newline = FALSE;

      if (( p = get_header_content( gfd->response.header, "Connection" ))
          != NULL && !strcmp( p, "close" )) {
        /* NTLM will be the death of me */
        int sock;

        gaim_debug_misc( __FUNCTION__,
                         "stupid NTLM closed connection. reopening\n" );

        if (( sock = gaim_proxy_connect( ac, gfd->website.address,
                                         gfd->website.port, url_fetched_cb,
                                         gfd )) < 0 ) {
          rvp_async_data( gc, gfd, 0 );
          destroy_fetch_url_data( gfd );
        } else {
          g_free( gfd->response.webdata );
          gfd->response.webdata = NULL;
          gfd->preserve = TRUE;
        }
      } else {
        gfd->resend = TRUE;
      }
      g_free( gfd->response.header );
      gfd->response.header = NULL;
      return;
    }
  }

  p = get_header_content( gfd->response.header, "Content-Type" );
  if ( p && !strcmp( p, "text/xml" )) {
    is_xml = TRUE;
  }

  /* decode the buffer - this has been determined by trial and error
     to be approximately correct, albeit stupid */
  if ( size > 0 ) {
    gchar *converted;
    GError *error = NULL;
    gsize read, written;

    converted = g_convert( gfd->response.webdata, size, "UCS-2LE" /* to */,
                           "UTF-8" /* from */, &read, &written, &error );
    if ( converted ) {
      gsize i;
      gfd->response.webdata = g_realloc( gfd->response.webdata, written / 2 );
      size = written / 2;
      for ( i = 0; i < written; i+= 2 ) {
        gfd->response.webdata[ i / 2 ] = converted[ i ];
      }
      g_free( converted );

      /* also update the gfud structure */
      gfd->len = written / 2;
    }
  }

  /*
   * if code is unset then we're handling an async message, which has
   * to be carrying an XML payload.
   */
  if ( code == 0 && is_xml ) {
    gint rc;
    /*
     * parse an async notification (e.g. IM, status change)
     */
    rc = parse_notify( gc, gfd );

    /* for now, send a 200 response regardless fixme */
    if ( gfd->sock != -1 ) {
      gchar *reply;

      reply = g_strdup_printf( "HTTP/1.1 %d %s\r\n"
                               "RVP-Notifications-Version: 0.2\r\n\r\n",
                               rc,
                               rc == 200 ? "Successful" : "Parse Error" );
      write( gfd->sock, reply, strlen( reply ));
      g_free( reply );
    }
  } else {
    /* if we're not connected, keep ticking the powerbar FIXME this is
       far too messy */
    if ( !GAIM_CONNECTION_IS_CONNECTED( gc ) &&
         code >= 200 &&
         code < 300
         && rd != NULL &&
         rd->login_step < MAX_LOGIN_STEPS ) {
      rd->login_step++;
    }

    switch ( code ) {
    case 200:
      if ( !strcmp( gfd->method, "SUBSCRIBE" )) {
        /* I should probably double-check, but a 200 response to a
           subscribe means it's a self-subscribe, i.e. a login or
           refresh */
        /* extract the useful info for this subscription */
        gchar *subs_id = get_header_content( gfd->response.header,
                                             "Subscription-Id" );
        gchar *lifetime = get_header_content( gfd->response.header,
                                              "Subscription-Lifetime" );


        rd->login_flags |= RVP_LOGIN_SUBSCRIBE;

        if ( subs_id != NULL ) {
          rd->subs_id = atoi( subs_id );
        }
        if ( lifetime != NULL ) {
          time_t timeout = atol( lifetime );
          if ( timeout == 0 ) {
            gaim_debug_error( __FUNCTION__, "can't parse timeout %s\n",
                              lifetime );
          } else {
            gaim_debug_misc( __FUNCTION__,
                             "main subscription expires in %u seconts\n",
                             timeout );
            timeout = timeout + time( NULL );
            rd->me.expires = timeout;
          }
        }

        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gchar *cmd[2];

          gaim_connection_update_progress( gc, _("Subscribed"), rd->login_step,
                                           MAX_LOGIN_STEPS );

          rvp_send_request( gc, "PROPPATCH", NULL );

          cmd[0] = g_strdup( "d:displayname" );
          cmd[1] = NULL;
          rvp_send_request( gc, "PROPFIND", cmd );
          g_free( cmd[0] );

          rvp_send_request( gc, "SUBSCRIPTIONS", NULL );
        } else {
          /* reset timeout on subscription */
          gaim_debug_misc( __FUNCTION__, "got subs response\n" );
        }
      } else if ( !strcmp( gfd->method, "SUBSCRIPTIONS" )) {
        rd->login_flags |= RVP_LOGIN_SUBSCRIPTIONS;
        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gaim_connection_update_progress( gc, _("Fetched subscriptions"),
                                           rd->login_step, MAX_LOGIN_STEPS );


          /* parse the subscriptions response */
          parse_subscriptions( gc, gfd );

          /* send an ACL self-check */
          rvp_send_request( gc, "ACL", NULL );
        } else {
          /* unlikely... */
          gaim_debug_misc( __FUNCTION__, "got subscriptions response\n" );
        }
      } else if ( !strcmp( gfd->method, "ACL" )) {
        if ( gfd->request.length == 0 ) {
          rd->login_flags |= RVP_LOGIN_ACL;
        }
        parse_acls( gc, gfd );
      } else if ( !strcmp( gfd->method, "UNSUBSCRIBE" )) {
        /* only handles self-unsubs xxx */
        rd->me.unsubbed = TRUE;
        gaim_debug_misc( __FUNCTION__, "unsubscribed\n" );
      }
      break;

    case 207:
      /* has to be XML by definition */
      if ( !strcmp( gfd->method, "PROPFIND" )) {
        rd->login_flags |= RVP_LOGIN_PROPFIND;
      } else if ( !strcmp( gfd->method, "PROPPATCH" )) {
        rd->login_flags |= RVP_LOGIN_PROPPATCH;
      }

      if ( !is_xml ) {
        gaim_connection_error( gc, _("Unparseable reply from server" ));
      } else {
        parse_multistatus( gc, gfd );
      }
      break;

    case 400:
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_connection_error( gc, _("Bad request"));
        gc->wants_to_die = TRUE;
      } else {
        /* hmm. should be able to recover from this. fixme. */
      }
      break;

    case 403:
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_connection_error( gc, _("Permission denied"));
        gc->wants_to_die = TRUE;
      } else {
        /* ? */
      }
      break;

    case 404:
      /* treat this as a 412 */
    case 412:
      if ( !strcmp( gfd->method, "NOTIFY" )) {
        GaimConversation *conv = NULL;
        gchar *recip;
        gchar *msg =
          g_strdup_printf( "%s\n<span color='red'>%s</span>",
                           "The following message could not be delivered to "
                           "all recipients:",
                           gfd->message == NULL ?
                           "[waider will eventually fix this]" :
                           gfd->message );

        /* a bit clunky */
        recip = rindex( gfd->website.page, '/' );
        if ( recip != NULL ) {
          recip = g_strdup( &recip[1] );
          conv = gaim_find_conversation_with_account( recip, ac );

          if ( conv != NULL && !g_strstr_len( gfd->request.webdata,
                                              gfd->request.length,
                                              "\r\nTypingUser: " )) {
            gaim_conversation_write( conv, NULL, msg,
                                     GAIM_MESSAGE_NO_LOG|GAIM_MESSAGE_ERROR,
                                     time( NULL ));
          }
          g_free( recip );
        }

        g_free( msg );
      } else {
        gaim_debug_error( __FUNCTION__, "%d from %s /%s\n", code,
                          gfd->method, gfd->website.page );
      }
      break;

    case 500:
    case 501: /* this is from my own server */
      if ( !strcmp( gfd->method, "NOTIFY" )) {
        /* this happens. the message has probably been
           delivered. Microsoft = teh sux0r */
        /* the docs say:
           This code indicates the absence of a PRINCIPAL from a
           session. For example, when a PRINCIPAL has left a
           discussion with multiple PRINCIPALS and its INSTANT INBOX
           receives notification of the session, it uses this return
           code. The sender of the INSTANT MESSAGE is then able to
           indicate that the PRINCIPAL has left the conversation. */

      } else {
        gaim_connection_error( gc, _("Internal server error" ));
        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gc->wants_to_die = TRUE;
        }
      }
      break;

    default:
      gaim_debug_misc( __FUNCTION__, "don't know what to do with %d\n", code );
      break;
    }

    if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
      if ( rd->login_flags == RVP_LOGIN_ALL ) {
        gchar *args[2];

        gaim_connection_set_state( gc, GAIM_CONNECTED );
        serv_finish_login( gc );

        /* send a self-subscribe to update/propchange as well as
           pragma/notify */
        args[0] = ( gchar *)rd->me.buddy;
        args[1] = NULL;
        rvp_send_request( gc, "SUBSCRIBE", args );
      }
    }
  }
  /* gfud code will clean up for us */
}

static void rvp_send_request( GaimConnection *gc, gchar *cmd, gchar **args ) {
  RVPData *rd = gc->proto_data;
  GaimFetchUrlData *gfud = NULL;
  GaimAccount *ac = gaim_connection_get_account( gc );
  xmlChar *content = NULL;
  gint content_length = 0;
  gchar *headers = NULL;
  gchar *url = NULL;
  int sock;

  /* may also need to allow propatch */
  if ( gc->wants_to_die && strcmp( cmd, "UNSUBSCRIBE" )) {
    gaim_debug_misc( __FUNCTION__, "We're dying. Not sending your %s\n", cmd );
    return;
  }

  /* client login/resubscribe */
  if ( !strcmp( cmd, "SUBSCRIBE" )) {
    if ( args == NULL ) {
      url = g_strdup( rd->me.principal );
      if ( rd->subs_id ) {
        headers = g_strdup_printf( "Subscription-Lifetime: %d\r\n"
                                   "Subscription-Id: %d\r\n"
                                   "RVP-From-Principal: %s\r\n",
                                   RVP_SUBS_TIME,
                                   rd->subs_id,
                                   rd->me.principal );
      } else {
        headers = g_strdup_printf( "Call-Back: %s\r\n"
                                   "Notification-Type: pragma/notify\r\n"
                                   "RVP-From-Principal: %s\r\n"
                                   "Subscription-Lifetime: %d\r\n",
                                   rd->client_url, rd->me.principal,
                                   RVP_SUBS_TIME );
      }
      /* no content */
    } else {
      GaimBuddy *buddy = (GaimBuddy *)args[0];
      RVPBuddy *rb = buddy->proto_data;

      url = g_strdup( rb->principal );
      if ( rb->subs_id ) {
        headers = g_strdup_printf( "Subscription-Id: %d\r\n"
                                   "Subscription-Lifetime: %d\r\n"
                                   "RVP-From-Principal: %s\r\n",
                                   rb->subs_id,
                                   RVP_SUBS_TIME,
                                   rd->me.principal );
      } else {
        headers = g_strdup_printf( "Notification-Type: update/propchange\r\n"
                                   "RVP-From-Principal: %s\r\n"
                                   "Subscription-Lifetime: %d\r\n"
                                   "Call-Back: %s\r\n",
                                   rd->me.principal,
                                   RVP_SUBS_TIME,
                                   rd->me.principal );
      }
    }
  } else if ( !strcmp( cmd, "PROPPATCH" )) {
    xmlDocPtr doc = xmlNewDoc( BAD_CAST "1.0" );
    xmlNodePtr root = xmlNewNode( NULL, BAD_CAST "d:propertyupdate" );
    xmlNodePtr node1, node2, prop;
    gchar *state;
    gchar *view_timeout;

    if ( args != NULL ) {
      state = g_strdup( (gchar *)args[0] );
    } else {
      if ( rd->me.state != NULL ) {
        state = g_strdup( rd->me.state );
      } else {
        state = g_strdup( "online" );
      }
    }

    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    prop = xmlNewNode( NULL, BAD_CAST "r:leased-value" );

    node1 = xmlNewNode( NULL, BAD_CAST "r:value" );
    node2 = xmlNewNode( NULL, BAD_CAST state );
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "r:default-value" );
    node2 = xmlNewNode( NULL, BAD_CAST "r:offline" );
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );

    view_timeout = g_strdup_printf( "%d", RVP_VIEW_TIME );
    node1 = xmlNewNode( NULL, BAD_CAST "d:timeout" );
    node2 = xmlNewText( BAD_CAST view_timeout );
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );
    g_free( view_timeout );

    node1 = xmlNewNode( NULL, BAD_CAST "r:state" );
    xmlAddChild( node1, prop );

    if ( rd->view_id ) {
      gchar *viewid = g_strdup_printf( "%d", rd->view_id );
      prop = xmlNewNode( NULL, BAD_CAST "r:view-id" );
      node2 = xmlNewText( BAD_CAST viewid );
      xmlAddChild( prop, node2 );
      xmlAddChild( node1, prop );
    }

    node2 = xmlNewNode( NULL, BAD_CAST "d:prop" );
    xmlAddChild( node2, node1 );
    node1 = xmlNewNode( NULL, BAD_CAST "d:set" );
    xmlAddChild( node1, node2 );
    xmlAddChild( root, node1 );

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    headers = g_strdup_printf( "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );

    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "PROPFIND" )) {
    gint propnum;
    xmlDocPtr doc = xmlNewDoc( BAD_CAST "1.0" );
    xmlNodePtr root = xmlNewNode( NULL, BAD_CAST "d:propfind" );

    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    /* practically speaking this is overkill */
    for ( propnum = 0; args[propnum]; propnum++ ) {
      xmlNodePtr prop = xmlNewNode( NULL, BAD_CAST "d:prop" );
      xmlNodePtr propname = xmlNewNode( NULL, BAD_CAST args[propnum] );
      xmlAddChild( prop, propname );
      xmlAddChild( root, prop );
    }

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    headers = g_strdup_printf( "Depth: 0\r\n" /* always 0 */
                               "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );

    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "SUBSCRIPTIONS" )) {
    headers = g_strdup_printf( "Notification-Type: update/propchange\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );
    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "ACL" )) {
    headers = g_strdup_printf( "RVP-From-Principal: %s\r\n",
                               rd->me.principal );
    if ( args != NULL ) {
      /* I think this is wrong */
      /* url = g_strdup( args[0] ); */
      /* and this is right */
      url = g_strdup( rd->me.principal );
      content = (xmlChar *)args[1];
      content_length = (gint)args[2];
    } else {
      url = g_strdup( rd->me.principal );
    }
  } else if ( !strcmp( cmd, "NOTIFY" )) {
    headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                               "Content-Type: text/xml\r\n"
                               "RVP-Hop-Count: 1\r\n"
                               "RVP-Ack-Type: DeepOr\r\n",
                               rd->me.principal );

    url = (gchar *)args[0];
    content = (xmlChar *)args[1];
    content_length = (gint)args[2];
  } else if ( !strcmp( cmd, "UNSUBSCRIBE" )) {
    if ( args == NULL ) {
      /* self-unsubscribe */
      url = g_strdup( rd->me.principal );
      headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                                 "Subscription-Id: %d\r\n",
                                 rd->me.principal,
                                 rd->subs_id );
    } else {
      GaimBuddy *buddy = (GaimBuddy *)args[0];
      RVPBuddy *rbuddy = buddy->proto_data;
      if ( rbuddy == NULL ||
           rbuddy->subs_id == 0 ||
           rbuddy->principal == NULL ) {
        gaim_debug_error( __FUNCTION__,
                          "Buddy problem unsubbing\n" ); /* Xxx */
        rvp_dump_buddy( buddy );
        return;
      }

      url = g_strdup( rbuddy->principal );
      headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                                 "Subscription-Id: %d\r\n",
                                 rbuddy->principal,
                                 rbuddy->subs_id );
    }
  }

  if ( rd->auth_type == RVP_AUTH_NTLM ) {
    gchar *ntlm, *newheaders;
    gaim_debug_misc( __FUNCTION__, "Preemptively adding NTLM\n" );

    ntlm = get_ntlm_msg1( rd->domain, rd->client_host );
    newheaders = g_strdup_printf( "Authorization: %s\r\n%s", ntlm, headers );

    g_free( headers );
    headers = newheaders;
  }

  gaim_debug_misc( __FUNCTION__, "allocated header %p\n", headers );

  gfud = g_new0( GaimFetchUrlData, 1 );
  gfud->request.header = headers;
  gfud->request.webdata = (gchar *)content; /* xxx */
  gfud->request.length = content_length;
  gfud->method = g_strdup( cmd );
  gfud->callback = rvp_async_data;
  gfud->user_data = gc;
  gfud->url = url;
  gaim_url_parse( gfud->url, &gfud->website.address, &gfud->website.port,
                  &gfud->website.page, &gfud->website.user,
                  &gfud->website.passwd );

  if ( !strcmp( cmd, "NOTIFY" ) && args[3] != NULL ) {
    gfud->message = g_strdup( args[3] );
  }

  if (( sock =
        gaim_proxy_connect( ac, gfud->website.address,
                            gfud->website.port, url_fetched_cb, gfud )) < 0 ) {
    rvp_async_data( gc, gfud, 0 ); /* xxx */
    destroy_fetch_url_data( gfud );
  }
}

/*
 * attached to the listening port
 */
static void rvp_listener_callback( gpointer data, gint source,
                                   GaimInputCondition cond ) {
  GaimConnection *gc = data;
  GaimFetchUrlData *fetch = NULL;
  RVPData *rd = gc->proto_data;
  int sd;

  /* this might have happened because of a disconnection... */
  if ( rd == NULL ) {
    gaim_debug_misc( __FUNCTION__, "erk. rd is null\n" );
    return;
  }

  if (( sd = accept( rd->listener_fd, 0, 0 )) < 0 ) {
      perror( "Accept failed.\n" );
      /* fixme drop the listener and make a new one? */
      gc->wants_to_die = TRUE;
      return;
  }

  /* These are incoming connections that we only expect to read from */
  fetch = g_new0( GaimFetchUrlData, 1 );
  fetch->callback = rvp_async_data;
  fetch->user_data = gc;
  fetch->sentreq = TRUE;
  fetch->data_len = 4096;
  fetch->response.webdata = g_malloc( fetch->data_len );
  fetch->sock = sd;
  fetch->inpa = gaim_input_add( sd, GAIM_INPUT_READ, url_fetched_cb, fetch );
}

/*
 * This is invoked via Gaim's event loop from rvp_login
 */
static void rvp_login_connect( gpointer data, gint source,
                               GaimInputCondition cond ) {
  GaimConnection *gc = data;
  struct rvp_data *rd = gc->proto_data;
  char buf[RVP_BUF_SIZE];
  char hostname[1024];
  char *tmp;
  const char *tmp2;
  struct hostent *p;
  struct sockaddr_in addr;
  int myhostname_found = FALSE;

  gaim_debug_misc( __FUNCTION__,  "enter\n" );

  if ( rd->fd != source ) {
    gaim_debug_misc( __FUNCTION__, "source fd was %d, now %d\n",
                     rd->fd, source );
    rd->fd = source;
  }

  if ( rd->fd == -1 ) {
    gaim_connection_error(gc, _("Unable to connect"));
    gc->wants_to_die = TRUE;
    return;
  }

  /*
   * Set up the callback socket
   */
  rvp_setup_listener( gc );

  /* check for failure - fixme, have rvp_setup_listener return a
     useful rc */
  if ( gc->wants_to_die == TRUE ) {
    return;
  }

  /* connect the listener up to Gaim's event loop */
  gaim_debug_misc( __FUNCTION__, "listener hookup\n" );
  rd->linpa =
    gaim_input_add( rd->listener_fd, GAIM_INPUT_READ, rvp_listener_callback,
                    gc );

  /* Create a session ID */
  rd->session_id = rvp_get_sessid();

  /* get server and client info */
  /* FIXME why bother doing this? */
  strcpy( hostname, gaim_connection_get_account( gc )->username );
  tmp = strtok( hostname, "@");
  rd->user = g_strdup( tmp );

  strcpy( hostname, gaim_connection_get_account( gc )->username );

  if ( gethostname( hostname, sizeof( hostname ) - 1 ) != 0 ) {
    gaim_debug_error( __FUNCTION__, "gethostname( %s ) failed\n", hostname );
    abort();
  }

  hostname[ sizeof( hostname ) - 1 ] = 0;
  p = gethostbyname( hostname );
  if ( p == NULL ) {
    /* we'll have to try some other method of sorting it out */
    socklen_t namelen = sizeof( addr );

    gaim_debug_error( __FUNCTION__,
                      "can't convert %s with gethostbyname (%s)\n",
                      hostname, strerror( errno ));

    if ( getsockname( rd->fd, (struct sockaddr *)&addr, &namelen )
         != 0 ) {
      gaim_debug_error( __FUNCTION__,
                        "can't convert %s with getsockname (%s)\n",
                        hostname, strerror( errno ));
    }
  } else if ( p && p->h_addr_list && p->h_addr_list[0] ) {
    addr.sin_addr = *(struct in_addr *)(p->h_addr_list[0]);
  } else {
    perror( "totally bogus" );
    abort();
  }

  /*
   * Some systems pick a hostname which isn't in DNS.  So, we we'll
   * try to use the IP address as the most general solution, as long
   * as the IP address does not start with "127."
   *
   * FIXME this is also pretty bogus, I think.
   */
  strncpy(buf,inet_ntoa(addr.sin_addr),sizeof(buf));
  if ( strncmp(buf,"127.",4) != 0 ) {
    myhostname_found = TRUE;
    strncpy(hostname,buf,sizeof(hostname));
    rd->client_host = g_strdup(hostname);
  }

  /* Does hostname contain a period */
  if ( !myhostname_found && strchr(p->h_name, '.')) {
    /* Yes, use it */
    strncpy(hostname,p->h_name,sizeof(hostname));
    rd->client_host = g_strdup(tmp);

    myhostname_found = TRUE;
  } else if ( !myhostname_found) {
    /* No, search all aliases */
    int x;

    if (p->h_aliases) {
      for (x = 0; p->h_aliases[x]; ++x) {
        if (p->h_aliases[x] && strchr(p->h_aliases[x], '.') &&
            (!strncasecmp(p->h_aliases[x], p->h_name, strlen(p->h_name)))) {
          strncpy(hostname,p->h_aliases[x],sizeof(hostname));
          myhostname_found = TRUE;
          rd->client_host = g_strdup(hostname);

          break;
        }
      }
    }
  }
  hostname[sizeof(hostname) - 1] = 0;

  /* use the configured option */
  if ((( tmp2 =
         gaim_account_get_string( gaim_connection_get_account( gc ),
                                  "myhost", NULL )) != NULL ) &&
      ( strlen( tmp2 ) > 0 )) {
    strncpy( hostname, tmp2, sizeof( hostname ));
    rd->client_host = g_strdup( hostname );
    myhostname_found = TRUE;
  }

  /* XXX */
  if ( ! myhostname_found ) {
    fprintf(stderr,
            "Unable to find our fully qualified hostname or IP address. "
            "Sorry, I'm giving up.\n");
    abort();
  } else {
    gaim_debug_misc( __FUNCTION__, "determined my hostname as %s\n",
                     hostname );
  }

  rd->client_url = g_strdup_printf( "http://%s:%d", hostname,
                                    rd->listener_port );

  /*
   * push a request out
   */
  rvp_send_request( gc, "SUBSCRIBE", NULL );
  gaim_debug_misc( __FUNCTION__, "exit\n" );
}

/*
 * Initial entrypoint for establishing a connection
 *
 * invokes rvp_login_connect via Gaim's event loop.
 */
static void rvp_login( GaimAccount *account ) {
  GaimConnection *gc = gaim_account_get_connection( account );
  RVPData *rd;
  const char *host = NULL, *domain, *id;
  const char *username;
  gchar **split;
  gchar *msg,  *srvname;

  if ( gc->proto_data != NULL ) {
    g_free( gc->proto_data ); /* xxx */
  }
  gc->proto_data = g_new0( RVPData, 1 );
  rd = gc->proto_data;

  rd->login_step = 0;
  rd->me.buddy = g_new0( GaimBuddy, 1 );
  rd->me.buddy->proto_data = rd;

  /* The MSN code does this, but the problem is that it causes
     hyperlinks to be sent with <a href... wrapped around them, and
     brokets get entity-encoded. And the Official Client doesn't
     understand that sort of thing. */
  /*  gc->flags |= GAIM_CONNECTION_HTML | GAIM_CONNECTION_FORMATTING_WBFO |
    GAIM_CONNECTION_NO_BGCOLOR | GAIM_CONNECTION_NO_FONTSIZE |
    GAIM_CONNECTION_NO_URLDESC;*/

  username = gaim_account_get_username( account );
  split = g_strsplit( username, "@", 2 );

  if ( split == NULL ) {
    gc->wants_to_die = TRUE;
    gaim_connection_error( gc, _("Invalid username."));
    return;
  }

  if ( split[1] != NULL ) {
    host = split[1];
  } else {
    host = gaim_account_get_string( account, "host", NULL );
  }

  if (( host == NULL || ( strlen( host ) == 0 ))) {
    gc->wants_to_die = TRUE;
    gaim_connection_error( gc, _("No login host specified."));
    return;
  }

  srvname = g_strconcat( "_rvp._tcp.", host, NULL );
  rd->port = 80; /* default */

  msg = g_strdup_printf( _("Locating RVP server for %s"), host );
  gaim_debug_misc( __FUNCTION__, "%s\n", msg );
  g_free( msg );

  rd->me.service = gethostbysrv( srvname, rd->me.service );
  if ( rd->me.service->host != NULL && rd->me.service->host != NULL ) {
    host = rd->me.service->host;
    rd->port = rd->me.service->port;
  } else {
    /* if we don't find a srv record, host is whatever we found above.
       For completeness, we should gethostbyname() it; if it's not
       valid, the final attempt is to do a NetBIOS lookup on it, after
       which we bail out. FIXME.
    */
    gaim_debug_misc( __FUNCTION__, "No srv record, falling back on %s\n",
                     host );
  }
  g_free( srvname );

  if ( rd->port != 80 ) {
    rd->me.principal = g_strdup_printf( "http://%s:%d/instmsg/aliases/%s",
                                         host, rd->port, split[0] );
  } else {
    rd->me.principal = g_strdup_printf( "http://%s/instmsg/aliases/%s",
                                        host, split[0] );
  }
  rd->me.buddy->name = g_strconcat( split[0], "@", host, NULL );
  rd->host = g_strdup( host );

  g_strfreev( split );

  msg = g_strdup_printf( _("Connecting to %s:%d"), host, rd->port );
  gaim_connection_update_progress( gc, msg, rd->login_step, MAX_LOGIN_STEPS );
  gaim_debug_misc( __FUNCTION__, "%s\n", msg );
  g_free( msg );

  domain = gaim_account_get_string( account, "domain", NULL );
  id = gaim_account_get_string( account, "id", NULL );

  /* do the preferred_port dance */
  if ( gaim_prefs_get_bool( "/core/network/ports_range_use" )) {
    rd->port_low = gaim_prefs_get_int( "/core/network/ports_range_start" );
    rd->port_high = gaim_prefs_get_int( "/core/network/ports_range_end" );
    /* make sure they're the right way around */
    if ( rd->port_high > rd->port_low ) {
      int tmp = rd->port_low;
      rd->port_low = rd->port_high;
      rd->port_high = tmp;
    }
  }

  rd->domain = g_strdup( domain );
  rd->authid = g_strdup( id );

  rd->fd =
    gaim_proxy_connect( account, rd->host, rd->port, rvp_login_connect, gc );

  if ( rd->fd < 0 ) {
    gaim_connection_error(gc, _("Failed to connect to server."));
  }
}

static gboolean rvp_load( GaimPlugin *plugin ) {
  gaim_debug_misc( __FUNCTION__, "rvp_load\n" );

  /* initialise the random-number code */
  init_seed( 0 );

  return TRUE;
}

static gboolean rvp_unload( GaimPlugin *plugin ) {
  gaim_debug_misc( __FUNCTION__,  "rvp_unload\n" );
  return TRUE;
}

static GaimPluginPrefFrame *get_plugin_pref_frame( GaimPlugin *plugin ) {
  GaimPluginPrefFrame *frame;
  GaimPluginPref *ppref;

  frame = gaim_plugin_pref_frame_new();

  ppref = gaim_plugin_pref_new_with_label( _("RVP Plugin Version $Revision: 1.87 $"));
  gaim_plugin_pref_frame_add( frame, ppref );

  ppref =
    gaim_plugin_pref_new_with_name_and_label( "/plugins/prpl/rvp/"
                                              "conv_close_notice",
                                              _("Display conversation "
                                                "closed notices"));
  gaim_plugin_pref_frame_add( frame, ppref );

  ppref =
    gaim_plugin_pref_new_with_name_and_label( "/plugins/prpl/rvp/"
                                              "conv_timeout_notice",
                                              _("Display timeout notices"));
  gaim_plugin_pref_frame_add( frame, ppref );

  return frame;
}

static GaimPluginUiInfo prefs_info = {
  get_plugin_pref_frame,
};

static GaimPluginProtocolInfo prpl_info = {
  0, /* no options */
  NULL,                   /* user_splits */
  NULL,                   /* protocol_options */
  { "png", 0, 0, 96, 96, GAIM_ICON_SCALE_SEND },    /* icon_spec */
  rvp_list_icon,          /* list_icon */
  rvp_list_emblems,       /* list_emblems */
  rvp_status_text,        /* status_text */
  NULL,                   /* tooltip_text */        /* not useful */
  rvp_away_states,        /* away_states */
  rvp_buddy_menu,         /* blist_node_menu */
  NULL,                   /* chat_info */
  NULL,                   /* chat_info_defaults */

  rvp_login,              /* login */
  rvp_close,              /* close */
  rvp_send_im,            /* send_im */
  NULL,                   /* set_info */           /* PROPPATCH - permitted? */
  rvp_send_typing,        /* send_typing */
  rvp_get_info,           /* get_info */           /* PROPFIND */
  rvp_set_away,           /* set_away */
  rvp_set_idle,           /* set_idle */
  NULL,                   /* change_passwd */      /* not permitted? */
  rvp_add_buddy,          /* add_buddy */
  NULL,                   /* add_buddies */        /* no UI */
  rvp_rem_buddy,          /* remove_buddy */
  NULL,                   /* remove_buddies */     /* no UI */
  rvp_add_permit,         /* add_permit */
  rvp_add_deny,           /* add_deny */
  rvp_rem_permit,         /* rem_permit */
  rvp_rem_deny,           /* rem_deny */
  rvp_set_permit_deny,    /* set_permit_deny */
  NULL,                   /* warn */
  rvp_chat_join,          /* join_chat */
  NULL,                   /* reject chat invite */ /* not available? */
  NULL,                   /* get_chat_name */
  rvp_chat_invite,        /* chat_invite */
  rvp_chat_leave,         /* chat_leave */
  NULL,                   /* chat_whisper */
  rvp_chat_send,          /* chat_send */
  rvp_keepalive,          /* keepalive */
  NULL,                   /* register_user */      /* not permitted */
  NULL,                   /* get_cb_info */
  NULL,                   /* get_cb_away */
  NULL,                   /* alias_buddy */        /* not permitted */
  NULL /* rvp_group_buddy */,        /* group_buddy */
  NULL /* rvp_rename_group */,       /* rename_group */
  rvp_free_buddy,         /* buddy_free */
  NULL /* rvp_convo_closed */,       /* convo_closed */
  rvp_normalize,          /* normalize */
  NULL /* rvp_set_buddy_icon */,     /* set_buddy_icon */
  NULL /* rvp_remove_group */,       /* remove_group */
  NULL,                   /* get_cb_real_name */
  NULL,                   /* set_chat_topic */
  NULL,                   /* find_blist_chat */
  NULL,                   /* roomlist_get_list */
  NULL,                   /* roomlist_cancel */
  NULL,                   /* roomlist_expand_category */
#ifdef WAIDER_DOES_FILES_NOW
  rvp_can_receive_file,   /* can_receive_file */
  rvp_send_file,          /* send_file */
#else
  NULL,
  NULL,
#endif
};

static GaimPluginInfo info = {
  GAIM_PLUGIN_MAGIC,
  GAIM_MAJOR_VERSION,
  GAIM_MINOR_VERSION,
  GAIM_PLUGIN_PROTOCOL,                             /**< type           */
  NULL,                                             /**< ui_requirement */
  0,                                                /**< flags          */
  NULL,                                             /**< dependencies   */
  GAIM_PRIORITY_DEFAULT,                            /**< priority       */

  PROTO_RVP,                                        /**< id             */
  "RVP",                                            /**< name           */
  VERSION,                                          /**< version        */
                                                      /**  summary        */
  N_("RVP Protocol Plugin"),
  /**  description    */
  N_("RVP Protocol Plugin"),
  "Weihua Sun <weihua@lucent.com>, Waider <waider@waider.ie>", /**< author         */
  GAIM_WEBSITE,                                     /**< homepage       */

  rvp_load,                                         /**< load           */
  rvp_unload,                                       /**< unload         */
  NULL,                                             /**< destroy        */

  NULL,                                             /**< ui_info        */
  &prpl_info,                                       /**< extra_info     */
  &prefs_info,                                      /**< prefs_info     */
  NULL /* rvp_actions */
};

static void init_plugin( GaimPlugin *plugin ) {
  GaimAccountOption *option;
  gchar *authid = NULL, *fixedport = NULL, *myhost = NULL;
  gint port = 0;

  option = gaim_account_option_string_new(_("Auth Host"), "host", "" );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  option = gaim_account_option_string_new(_("Auth Domain"), "domain", "" );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  /* compatability with old code */
  authid = getenv( "RVP_AUTH_ID" );
  fixedport = getenv( "RVP_FIXED_PORT" );
  if ( fixedport != NULL ) {
    port = atoi( fixedport );
  }
  myhost = getenv( "GAIM_HOSTNAME" );

  option = gaim_account_option_string_new(_("Auth ID"), "id", authid );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  option = gaim_account_option_int_new(_("Fixed Port"), "port", port );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  option = gaim_account_option_string_new( _("My Hostname"), "myhost",
                                           myhost );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  gaim_prefs_add_none( "/plugins/prpl/rvp");
  /*
    these are taken from the Microsoft code, but are currently
    noops in this plugin
     gaim_prefs_add_bool("/plugins/prpl/rvp/conv_close_notice",   TRUE);
     gaim_prefs_add_bool("/plugins/prpl/rvp/conv_timeout_notice", TRUE);
  */
}

GAIM_INIT_PLUGIN(rvp, init_plugin, info);
